#!/usr/bin/env python

"""bulge_graph.py: A graph representation of RNA secondary structure based
   on its decomposition into primitive structure types: stems, hairpins,
   interior loops, multiloops, etc..."""

__author__      = "Peter Kerpedjiev"
__copyright__   = "Copyright 2012, 2013, 2014"
__version__     = "0.1"
__maintainer__  = "Peter Kerpedjiev"
__email__       = "pkerp@tbi.univie.ac.at"

import sys
import itertools as it
import collections as c
import math
import random
import itertools as it
import mirge.forgi.utilities.debug as fud
import mirge.forgi.utilities.stuff as cus
import mirge.forgi.threedee.utilities.mcannotate as ftum
import mirge.forgi.threedee.utilities.vector as cuv

def error_exit(message):
    print(message, file=sys.stderr)
    sys.exit(1)

# A wrapper for a simple dictionary addition
# Added so that debugging can be made easier
def add_bulge(bulges, bulge, context, message):
    #print >>sys.stderr,"Adding bulge", context, bulge, message
    #bulge = (context, bulge)
    bulges[context] = bulges.get(context, []) + [bulge]
    return bulges

def any_difference_of_one(stem, bulge):
    '''
    See if there's any difference of one between the two
    ends of the stem [(a,b),(c,d)] and a bulge (e,f)

    :param stem: A couple of couples (2 x 2-tuple) indicating the start and end
                 nucleotides of the stem in the form ((s1, e1), (s2, e2))
    :param bulge: A couple (2-tuple) indicating the first and last position
                  of the bulge.
    :return: True if there is an overlap between the stem nucleotides and the 
                  bulge nucleotides. False otherwise
    '''
    for stem_part in stem:
        for part in stem_part:
            for bulge_part in bulge:
                if abs(bulge_part - part) == 1:
                    return True
    return False

def print_bulges(bulges):
    '''
    Print the names and definitions of the bulges.

    :param bulges: A list of tuples of the form [(s, e)] where s and e are the 
                   numbers of the nucleotides at the start and end of the bulge.
    '''
    for i in range(len(bulges)):
            #print "bulge:", bulge
        bulge_str = "define b%d 1" % (i)
        bulge = bulges[i]
        bulge_str += " %d %d" % (bulge[0]+1, bulge[1]+1)
        print(bulge_str)

def condense_stem_pairs(stem_pairs):
    '''
    Given a list of stem pairs, condense them into stem definitions

    I.e. the pairs (0,10),(1,9),(2,8),(3,7) can be condensed into
    just the ends of the stem: [(0,10),(3,7)]

    :param stem_pairs: A list of tuples containing paired base numbers.

    :returns: A list of tuples of tuples of the form [((s1, e1), (s2, e2))]
                  where s1 and e1 are the nucleotides at one end of the stem
                  and s2 and e2 are the nucleotides at the other.
    '''
    stem_pairs.sort()

    prev_pair = (-10, -10)

    stems = []
    start_pair = None

    for pair in stem_pairs:
        # There's a potential bug here since we don't check the direction
        # but hopefully it won't bite us in the ass later
        if abs(pair[0] - prev_pair[0]) != 1 or abs(pair[1] - prev_pair[1]) != 1:
            if start_pair != None:
                stems += [(start_pair, prev_pair)]
            start_pair = pair
    
        prev_pair = pair

    if start_pair != None:
        stems += [(start_pair, prev_pair)]

    return stems

def print_brackets(brackets):
    '''
    Print the brackets and a numbering, for debugging purposes

    :param brackets: A string with the dotplot passed as input to this script.
    '''
    numbers = [chr(ord('0') + i % 10) for i in range(len(brackets))]
    tens = [chr(ord('0') + i / 10) for i in range(len(brackets))]
    print("brackets:\n", brackets, "\n", "".join(tens), "\n" ,"".join(numbers))

def find_bulges_and_stems(brackets):
    '''
    Iterate through the structure and enumerate the bulges and the stems that are
    present.

    The returned stems are of the form [[(s1, s2), (e1,e2)], [(s1,s2),(e1,e2)],...]
    where (s1,s2) are the residue numbers of one end of the stem and (e1,e2) are the
    residue numbers at the other end of the stem
    (see condense_stem_pairs)

    The returned bulges are of the form [(s,e), (s,e),...] where s is the start of a bulge
    and e is the end of a bulge

    :param brackets: A string with the dotplot passed as input to this script.
    '''
    prev = 'x'
    context = 0

    bulges = dict()
    finished_bulges = []
    context_depths = dict()

    opens = []
    stem_pairs = []

    stems = dict()

    dots_start = 0
    dots_end = 0

    context_depths[0] = 0

    i = 0
    for i in range(len(brackets)):
        if brackets[i] == '(':
            opens.append(i)

            if prev == '(':
                context_depths[context] = context_depths.get(context, 0) + 1
                continue
            else:
                context += 1
                context_depths[context] = 1

            if prev == '.':
                dots_end = i-1
                bulges = add_bulge(bulges, (dots_start, dots_end), context, "4")

        if brackets[i] == ')':
            if len(opens) == 0:
                error_exit("ERROR: Unmatched close bracket")

            stem_pairs.append((opens.pop(), i))

            context_depths[context] -= 1

            if context_depths[context] == 0:
                if context in bulges:
                    finished_bulges += bulges[context]
                bulges[context] = []
                context -= 1

            if prev == '.':
                dots_end = i-1
                bulges = add_bulge(bulges, (dots_start, dots_end), context, "2")

        if brackets[i] == '.':
            if prev == '.':
                continue

            dots_start = i

        prev = brackets[i]

    if prev == '.':
        dots_end = i
        bulges = add_bulge(bulges, (dots_start, dots_end), context, "7")
    elif prev == '(':
        print("Unmatched bracket at the end", file=sys.stderr)
        sys.exit(1)
    '''
    elif prev == ')':
        bulges = add_bulge(bulges, (i+1, i+1), context, "8")
    '''
    
    if context in bulges.keys():
        finished_bulges += bulges[context]

    if len(opens) > 0:
        error_exit("ERROR: Unmatched open bracket")

    stem_pairs.sort()
    stems = condense_stem_pairs(stem_pairs)
    
    return (finished_bulges, stems)

def print_name(filename):
    print("name", os.path.splitext(filename)[0])



class BulgeGraph(object):
    def __init__(self, bg_file=None, dotbracket_str='', seq=''):
        self.seq_length = 0
        self.ang_types = None
        self.mst = None
        self.build_order = None
        self.name = "untitled"
        self.defines = dict()
        self.edges = c.defaultdict(set)
        self.longrange = c.defaultdict(set)
        self.weights = dict()

        # sort the coordinate basis for each stem
        self.bases = dict()
        self.stem_invs = dict()
        self.seq_ids = []

        self.name_counter = 0

        if dotbracket_str != '':
            self.from_dotbracket(dotbracket_str)

        self.seq = seq
        for i, s in enumerate(seq):
            self.seq_ids += [(' ', str(i+1), ' ')]

        if bg_file is not None:
            self.from_bg_file(bg_file)

    # get an internal index for a named vertex
    # this applies to both stems and edges
    def get_vertex(self, name = None):
        '''
        Return a new unique vertex name.
        '''

        if name == None:
            name = "x%d" % (self.name_counter)
            self.name_counter += 1

        return name

    def element_length(self, key):
        '''
        Get the number of residues that are contained within this element.

        @param key: The name of the element.
        '''
        d = self.defines[key]
        length = 0

        for i in range(0, len(d), 2):
            length += d[i+1] - d[i] + 1

        return length 
            
    def stem_length(self, key):
        '''
        Get the length of a particular element. If it's a stem, it's equal to
        the number of paired bases. If it's an interior loop, it's equal to the
        number of unpaired bases on the strand with less unpaired bases. If
        it's a multiloop, then it's the number of unpaired bases.
        '''
        d = self.defines[key]
        if key[0] == 's' or key[0] == 'y':
            return (d[1] - d[0]) + 1
        elif key[0] == 'f':
            return self.get_bulge_dimensions(key)[0]
        elif key[0] == 't':
            return self.get_bulge_dimensions(key)[1]
        elif key[0] == 'h':
            return self.get_bulge_dimensions(key)[0]
        else:
            return min(self.get_bulge_dimensions(key))

    def get_single_define_str(self, key):
        '''
        Get a define string for a single key.
        '''
        return "define %s %s" % ( key, " ".join([str(d) for d in self.defines[key]]))

    def get_define_str(self):
        '''
        Convert the defines into a string. 

        Format:

        define [name] [start_res1] [end_res1] [start_res2] [end_res2]
        '''
        defines_str = ''

        # a method for sorting the defines
        def define_sorter(k):
            drni = self.define_residue_num_iterator(k, adjacent=True)
            return next(drni)

        for key in sorted(self.defines.keys(), key=define_sorter):
            defines_str += self.get_single_define_str(key)
            #defines_str += "define %s %s" % ( key, " ".join([str(d) for d in self.defines[key]]))
            defines_str += '\n'
        return defines_str

    def get_length_str(self):
        return "length " + str(self.seq_length) + '\n'

    def get_connect_str(self):
        '''
        Get the connections of the bulges in the graph.

        Format:

        connect [from] [to1] [to2] [to3]
        '''

        whole_str = ''
        for key in self.edges:
            if len(self.edges[key]) == 0:
                continue

            # Our graph will be defined by the stems and the bulges they connect to
            name = key
            if name[0] == 's':
                out_str = "connect %s" % (name)

                for dest in self.edges[key]:
                    out_str += " %s" % (dest)

                whole_str += out_str
                whole_str += '\n'

        return whole_str

    def get_sequence_str(self):
        '''
        Return the sequence along with its keyword. I.e.

            seq ACGGGCC
        '''
        if len(self.seq) > 0:
            return "seq %s\n" % (self.seq)
        else:
            return ""

    def get_seq_ids_str(self):
        '''
        Return the sequence id string

        seq_ids 1 2 2.A 17
        '''
        out_str = "seq_ids "
        out_str += " ".join(map(ftum.format_resid, self.seq_ids))
        out_str += "\n"

        return out_str

    def get_name_str(self):
        '''
        Return the name of this structure along with its keyword:

            name 1y26
        '''
        return "name %s\n" % (self.name)



    def to_bg_string(self):
        '''
        Output a string representation that can be stored and reloaded.

        '''
        out_str = ''
        out_str += self.get_name_str()
        out_str += self.get_length_str()
        out_str += self.get_sequence_str()
        out_str += self.get_seq_ids_str()
        out_str += self.get_define_str()
        out_str += self.get_connect_str()

        return out_str

    def to_file(self, filename):
        with open(filename, 'w') as f:
            out_str = self.to_bg_string()

            f.write(out_str)

    def to_element_string(self):
        '''
        Create a string similar to dotbracket notation that identifies what
        type of element is present at each location.

        For example the following dotbracket:

        ..((..))..

        Should yield the following element string:

        ffsshhsstt

        Indicating that it begins with a fiveprime region, continues with a
        stem, has a hairpin after the stem, the stem continues and it is terminated
        by a threeprime region.
        '''
        output_str = [' ' for i in range(self.seq_length+1)]

        for d in self.defines.keys():
            for resi in self.define_residue_num_iterator(d, adjacent=False):
                output_str[resi] = d[0]

        return "".join(output_str).strip()

    def define_range_iterator(self, node, adjacent=False, seq_ids=False):
        '''
        Return the ranges of the nucleotides in the define.

        In other words, if a define contains the following: [1,2,7,8]
        The ranges will be [1,2] and [7,8].

        @param adjacent: Use the nucleotides in the neighboring element which
                         connect to this element as the range starts and ends.
        @return: A list of two-element lists
        '''
        a = iter(self.defines[node])
        ranges = zip(a,a)

        if node[0] == 'i':
            # interior loops have to be treated specially because
            # they might have a bulge that has no unpaired nucleotides on one strand

            if adjacent:
                conns = self.connections(node)
                s1 = self.defines[conns[0]]
                s2 = self.defines[conns[1]]

                if adjacent:
                    # offset by one, which will be reversed in the yield step
                    # below
                    ranges =  [[s1[1]+1, s2[0]-1], [s2[3]+1, s1[2]-1]]

        if node[0] == 'm':
            if adjacent:
                conns = self.connections(node)
                s1 = self.get_sides_plus(conns[0], node)[0]
                s2 = self.get_sides_plus(conns[1], node)[0]

                rnge = sorted([self.defines[conns[0]][s1],
                                self.defines[conns[1]][s2]])
                ranges = [[rnge[0] + 1, rnge[1] - 1]]

        for (ds1, ds2) in ranges:
            if adjacent:
                if ds1 > 1:
                    ds1 -= 1
                if ds2 < self.seq_length-1:
                    ds2 += 1

            if seq_ids:
                # this will cause problems if the nucleotide has insertion codes
                yield [self.seq_ids[ds1-1],self.seq_ids[ds2-1]]
            else:
                yield [ds1,ds2]


    def define_residue_num_iterator(self, node, adjacent=False, seq_ids=False):
        '''
        Iterate over the residue numbers that belong to this node.

        :param node: The name of the node
        '''
        for r in self.define_range_iterator(node, adjacent, seq_ids=False):
            for i in range(r[0], r[1] + 1):
                if seq_ids:
                    yield self.seq_ids[i-1]
                else:
                    yield i

    def iterate_over_seqid_range(self, start_id, end_id):
        '''
        Iterate over the seq_ids between the start_id and end_id.
        '''
        i1 = self.seq_ids.index(start_id)
        i2 = self.seq_ids.index(end_id)

        for i in range(i1, i2+1):
            yield self.seq_ids[i]

    def create_bulge_graph(self, stems, bulges):
        '''
        Find out which stems connect to which bulges

        Stems and bulges which share a nucleotide are considered connected.

        :param stems: A list of tuples of tuples of the form [((s1, e1), (s2, e2))]
                      where s1 and e1 are the nucleotides at one end of the stem
                      and s2 and e2 are the nucleotides at the other.

        :param bulges: A list of tuples of the form [(s, e)] where s and e are the 
                       numbers of the nucleotides at the start and end of the bulge.
        '''
        for i in range(len(stems)):
            stem = stems[i]
            for j in range(len(bulges)):
                bulge = bulges[j]
                if any_difference_of_one(stem, bulge):
                    self.edges['y%d' % (i)].add('b%d' % (j))
                    self.edges['b%d' % (j)].add('y%d' % (i))

    def create_stem_graph(self, stems, bulge_counter):
        '''
        Determine which stems are connected to each other. A stem can be connected to
        another stem when there is an interior loop with an unpaired nucleotide on
        one side. In this case, a bulge will be created on the other side, but it
        will only consist of the two paired bases around where the unpaired base 
        would be if it existed.

        The defines for these bulges will be printed as well as the connection strings
        for the stems they are connected to.

        :param stems: A list of tuples of tuples of the form [((s1, e1), (s2, e2))]
                      where s1 and e1 are the nucleotides at one end of the stem
                      and s2 and e2 are the nucleotides at the other.
        :param bulge_counter: The number of bulges that have been encountered so far.

        :returns: A dictionary indexed by the number of a stem, containing a set of the 
                 other stems that the index is connected to.
        '''
        #print "stems:", stems
        stem_stems = dict()
        define_text = ""
        connect_text = ""
        for i in range(len(stems)):
            stem = stems[i]
            for j in range(i+1, len(stems)):
                for k1 in range(2):
                    # don't fear the for loop
                    for k2 in range(2):
                        for l1 in range(2):
                            for l2 in range(2):
                                s1 = stems[i][k1][l1]
                                s2 = stems[j][k2][l2]
                                if abs(s1 - s2) == 1:
                                    stem_stems_set = stem_stems.get(i, set())
                                    if j not in stem_stems_set:
                                        bn = 'b%d' % (bulge_counter)
                                        #self.defines[bn] = [min(s1, s2)+1, max(s1, s2)+1]
                                        self.defines[bn] = []
                                        self.weights[bn] = 1

                                        self.edges['y%d' % i].add(bn)
                                        self.edges[bn].add('y%d' % i)

                                        self.edges['y%d' % j].add(bn)
                                        self.edges[bn].add('y%d' % j)

                                        bulge_counter += 1
                                        stem_stems_set.add(j)
                                    stem_stems[i] = stem_stems_set

        for d in self.defines.keys():
            if d[0] != 'y':
                continue

            (s1, e1, s2, e2) = self.defines[d] 
            if abs(s2 - e1) == 1:
                bn = 'b%d' % (bulge_counter)

                self.defines[bn] = []
                self.weights[bn] = 1

                self.edges[bn].add(d)
                self.edges[d].add(bn)

                bulge_counter += 1
            
        return stem_stems

    def remove_vertex(self, v):
        '''
        Delete a node after merging it with another

        :param v: The name of the node
        '''
        # delete all edges to this node
        for key in self.edges[v]:
            self.edges[key].remove(v)

        for edge in self.edges:
            if v in self.edges[edge]:
                self.edges[edge].remove(v)

        # delete all edges from this node
        del self.edges[v]
        del self.defines[v]

    def reduce_defines(self):
        """
        Make defines like this:

        define x0 2 124 124 3 4 125 127 5 5
        
        Into this:

        define x0 2 3 5 124 127

        That is, consolidate contiguous bulge region defines.
        """

        new_defines = []

        for key in self.defines.keys():
            if key[0] != 's':
                assert(len(self.defines[key]) % 2 == 0)

                j = 0
                new_j = 0

                while new_j < len(self.defines[key]):

                    j = new_j
                    new_j += j + 2

                    (f1, t1) = (int(self.defines[key][j]), int(self.defines[key][j+1]))

                    # remove bulges of length 0
                    if f1 == -1 and t1 == -2:
                        del self.defines[key][j]
                        del self.defines[key][j]
                        
                        new_j = 0
                        continue

                    # merge contiguous bulge regions
                    for k in range(j+2, len(self.defines[key]), 2):
                        if key[0] == 'y':
                            # we can have stems with defines like: [1,2,3,4]
                            # which would imply a non-existant loop at its end
                            continue

                        (f2, t2) = (int(self.defines[key][k]), int(self.defines[key][k+1]))


                        if t2 + 1 != f1 and t1 + 1 != f2:
                            continue

                        if t2 + 1 == f1:
                            self.defines[key][j] = str(f2)
                            self.defines[key][j+1] = str(t1)
                        elif t1 + 1 == f2:
                            self.defines[key][j] = str(f1)
                            self.defines[key][j+1] = str(t2)

                        del self.defines[key][k]
                        del self.defines[key][k]

                        new_j = 0

                        break

    def merge_vertices(self, vertices):
        '''
        This is done when two of the outgoing strands of a stem
        go to different bulges
        It is assumed that the two ends are on the same sides because
        at least one vertex has a weight of 2, implying that it accounts
        for all of the edges going out of one side of the stem

        :param vertices: A list of vertex names to combine into one.
        '''
        merge_str = ""
        new_vertex = self.get_vertex()
        self.weights[new_vertex] = 0

        #assert(len(vertices) == 2)

        connections = set()
        needs_merging = set()

        for v in vertices:
            merge_str += " %s" % (v)

            # what are we gonna merge?
            for item in self.edges[v]:
                connections.add(item)

            # Add the definition of this vertex to the new vertex
            #self.merge_defs[new_vertex] = self.merge_defs.get(new_vertex, []) + [v]

            if v[0] == 's':
                self.defines[new_vertex] = self.defines.get(new_vertex, []) + [self.defines[v][0], self.defines[v][2]] + [self.defines[v][1], self.defines[v][3]]
            else:
                self.defines[new_vertex] = self.defines.get(new_vertex,[]) + self.defines[v]


            self.weights[new_vertex] += 1

            # remove the old vertex, since it's been replaced by new_vertex
            self.remove_vertex(v)
            self.reduce_defines()

        #self.weights[new_vertex] = 2
        for connection in connections:
            self.edges[new_vertex].add(connection)
            self.edges[connection].add(new_vertex)

        return new_vertex

    def find_bulge_loop(self, vertex, max_length=4):
        '''
        Find a set of nodes that form a loop containing the
        given vertex and being no greater than 4 nodes long.

        :param vertex: The vertex to start the search from.
        :returns: A list of the nodes in the loop.
        '''
        visited = set()
        to_visit = [(key, 0) for key in self.edges[vertex]]
        visited.add(vertex)
        in_path = [vertex]

        while len(to_visit) > 0:
            (current, depth) = to_visit.pop()
            visited.add(current)

            in_path = in_path[:depth]
            in_path.append(current)

            for key in self.edges[current]:
                if key == vertex and depth > 1:
                    if len(in_path[:depth+1]) > max_length:
                        continue
                    else:
                        return in_path[:depth+1]
                
                if key not in visited:
                    to_visit.append((key, depth+1))
        return []

    def add_node(self, name, edges, define, weight=1):
        self.defines[name] = define
        self.edges[name] = edges
        self.weights[name] = weight

        for edge in self.edges[name]:
            self.edges[edge].add(name)

    def dissolve_stem(self, key):
        '''
        Remove a stem. This means that we need
        to reconfigure all of the adjacent elements in such a manner
        that they now include the nucleotides that were formerly 
        in this stem.
        '''
        connections = self.edges[key]
        all_defines = []

        # get all the unpaired regions
        for c in list(connections) + [key]:
            for x in cus.grouped(self.defines[c], 2):
                all_defines += [x]

        # condense them into contiguous regions of unpaired bases
        intervals = cus.merge_intervals(all_defines, diff=1)

        # remove the stem
        self.remove_vertex(key)

        to_remove = set()
        # find out which elements connect to each of the new
        # contiguous regions
        for i in intervals:
            new_connections = set()
            for c in connections:
                # get the connections of the nodes to be removed
                for nc in self.edges[c]:
                    if nc == i:
                        continue
                    for cd in self.defines[nc]:
                        for id in i:
                            if abs(cd - id) == 1:
                                new_connections.add(nc)
                to_remove.add(c)
            
            # add a new node corresponding to this unpaired region
            self.add_node(self.get_vertex(), new_connections, i, 1)

        # remove the vertices that have been consolidated
        for r in to_remove:
            self.remove_vertex(r)

        self.relabel_nodes()

    def collapse(self):
        '''
        If any vertices form a loop, then they are either a bulge region of 
        a fork region. The bulge (interior loop) regions will be condensed 
        into one node.
        '''

        new_vertex = True
        while new_vertex:
            new_vertex = False
            bulges = [k for k in self.defines if k[0] != 'y']

            for (b1, b2) in it.combinations(bulges, r=2):
                if self.edges[b1] == self.edges[b2] and len(self.edges[b1]) > 1:
                    elist = list(self.edges[b1])
                    connections = self.connections(b1)

                    all_connections = [sorted((self.get_sides_plus(connections[0], b1)[0],
                                              self.get_sides_plus(connections[0], b2)[0])),
                                       sorted((self.get_sides_plus(connections[1], b1)[0],
                                              self.get_sides_plus(connections[1], b2)[0]))]

                    if all_connections == [[1,2],[0,3]]:
                        # interior loop
                        self.merge_vertices([b1,b2])
                        new_vertex = True
                        break

    def interior_loop_iterator(self):
        """
        Iterate over all of the interior loops.

        An interior loop can only have two connections: to the two stems which it links. 
        """

        for key in self.defines.keys():
            if key[0] == 'i':
                yield key

    def relabel_node(self, old_name, new_name):
        '''
        Change the name of a node.

        param old_name: The previous name of the node
        param new_name: The new name of the node
        '''
        #replace the define name
        define = self.defines[old_name]

        del self.defines[old_name]
        self.defines[new_name] = define

        #replace the index into the edges array
        edge = self.edges[old_name]
        del self.edges[old_name]
        self.edges[new_name] = edge

        #replace the name of any edge that pointed to old_name
        for k in self.edges.keys():
            new_edges = set()
            for e in self.edges[k]:
                if e == old_name:
                    new_edges.add(new_name)
                else:
                    new_edges.add(e)
            self.edges[k] = new_edges

    def compare_stems(self, b):
        '''
        A function that can be passed in as the key to a sort.
        '''
        return (self.defines[b][0], 0)


    def compare_bulges(self, b):
        connections = self.connections(b)

        return (self.defines[connections[0]][0], 
                self.defines[connections[1]][0])

    def compare_hairpins(self, b):
        connections = self.connections(b)

        return (self.defines[connections[0]][1], sys.maxsize)

    def relabel_nodes(self):
        '''
        Change the labels of the nodes to be more indicative of their nature.

        s: stem
        h: hairpin
        i: interior loop
        m: multiloop
        f: five-prime unpaired
        t: three-prime unpaired
        '''
        stems = []
        hairpins = []
        interior_loops = []
        multiloops = []
        fiveprimes = []
        threeprimes = []

        for d in self.defines.keys():
            if d[0] == 'y' or d[0] == 's':
                stems += [d]

                stems.sort(key=self.compare_stems)
                continue

            if len(self.defines[d]) == 0 and len(self.edges[d]) == 1:
                hairpins += [d]
                continue

            if len(self.defines[d]) == 0 and len(self.edges[d]) == 2:
                multiloops += [d]
                continue

            if len(self.edges[d]) <= 1 and self.defines[d][0] == 1:
                fiveprimes += [d]
                continue

            if len(self.edges[d]) == 1 and self.defines[d][1] == self.seq_length:
                threeprimes += [d]
                continue

            if (len(self.edges[d]) == 1 and 
                self.defines[d][0] != 1 and 
                self.defines[d][1] != self.seq_length):
                hairpins += [d]

                hairpins.sort(key=self.compare_hairpins)
                continue

            if d[0] == 'm' or (d[0] != 'i' and len(self.edges[d]) == 2 and 
                self.weights[d] == 1 and 
                self.defines[d][0] != 1 and
                self.defines[d][1] != self.seq_length):
                multiloops += [d]

                multiloops.sort(key=self.compare_bulges)
                continue

            if d[0] == 'i' or self.weights[d] == 2:
                interior_loops += [d]
                interior_loops.sort(key=self.compare_stems)

        for d in fiveprimes:
            self.relabel_node(d, 'f1')
        for d in threeprimes:
            self.relabel_node(d, 't1')

        for i,d in enumerate(stems):
            self.relabel_node(d, 's%d' % (i))
        for i,d in enumerate(interior_loops):
            self.relabel_node(d, 'i%d' % (i))
        for i,d in enumerate(multiloops):
            self.relabel_node(d, 'm%d' % (i))
        for i,d in enumerate(hairpins):
            self.relabel_node(d, 'h%d' % (i))

    def has_connection(self, v1, v2):
        """ Is there an edge between these two nodes """

        if v2 in self.edges[v1]:
            return True
        else:
            # two multiloops can be connected at the end of a stem
            for e in self.edges[v1]:
                if e[0] != 's':
                    continue

                if v2 in self.edges[e]:
                    (s1b, s1e) = self.get_sides(e, v1)
                    (s2b, s2e) = self.get_sides(e, v2)

                    if s1b == s2b:
                        return True

            return False

    def connection_type(self, define, connections):
        '''
        Classify the way that two stems are connected according to the type
        of bulge that separates them.

        Potential angle types for single stranded segments, and the ends of
        the stems they connect:

            1   2 (1, 1) #pseudoknot
            1   0 (1, 0)
            3   2 (0, 1)
            3   0 (0, 0)

        @param define: The name of the bulge separating the two stems
        @param connections: The two stems and their separation
        '''

        if define[0] == 'i':
            # interior loop, we just have to check if 
            # connections[0] < connections[1]
            if self.defines[connections[0]][0] < self.defines[connections[1]][0]:
                return 1
            else:
                return -1
        elif define[0] == 'm':
            (s1c, b1c) = self.get_sides_plus(connections[0], define)
            (s2c, b2c) = self.get_sides_plus(connections[1], define)

            if (s1c, s2c) == (1, 0):
                return 2
            elif (s1c, s2c) == (0, 1):
                return -2
            elif (s1c, s2c) == (3, 0):
                return 3
            elif (s1c, s2c) == (0, 3):
                return -3
            elif (s1c, s2c) == (2, 3):
                return 4
            elif (s1c, s2c) == (3, 2):
                return -4

            # the next two refer to pseudoknots
            elif (s1c, s2c) == (2, 1):
                return 5
            elif (s1c, s2c) == (1, 2):
                return -5
            else:
                raise Exception("Weird angle type: (s1c, s2c) = (%d, %d)" % 
                                (s1c, s2c))
        else:
            raise Exception("connection_type called on non-interior loop/multiloop")

    def connection_ends(self, connection_type):
        """
        Find out which ends of the stems are connected by a particular angle
        type.

        @param connection_type: The angle type, as determined by which corners
                                of a stem are connected
        @return: (s1e, s2b)
        """
        ends = ()

        if abs(connection_type) == 1:
            ends = (1, 0)
        elif abs(connection_type) == 2:
            ends = (1, 0)
        elif abs(connection_type) == 3:
            ends = (0, 0)
        elif abs(connection_type) == 4:
            ends = (1, 0)
        elif abs(connection_type) == 5:
            ends = (1, 1)
        else:
            raise Exception('Unknown connection type: %d' % (connection_type))

        if connection_type < 0:
            return ends[::-1]
        else:
            return ends

    def find_multiloop_loops(self):
        '''
        Find out which defines are connected in a multiloop.

        @return: A list of sets, where each set contains the names of
                 the elements in a particular multiloop.
        '''
        multis = []
        visited = set()
        for d in self.mloop_iterator():
            if d in visited:
                continue
            # use a really high loop length
            v = self.find_bulge_loop(d , max_length=400)
            v += [d]
            multis += [set(v)]

            for d in v:
                visited.add(d)

        return multis

    def from_fasta(self, fasta_str, dissolve_length_one_stems=False):
        '''
        Create a bulge graph from a fasta-type file containing the following
        format:

            > id
            ACCGGGG
            ((...))
        '''
        lines = fasta_str.split('\n')
        self.from_dotbracket(lines[2].strip(), dissolve_length_one_stems)
        self.name = lines[0].strip('>')
        self.seq = lines[1].strip()

        self.seq_ids_from_seq()

    def seq_ids_from_seq(self):
        '''
        Get the sequence ids of the string.
        '''
        self.seq_ids = []

        # when provided with just a sequence, we presume that the
        # residue ids are numbered from 1-up
        for i, s in enumerate(self.seq):
            self.seq_ids += [(' ', i+1, ' ')]

    def remove_degenerate_nodes(self):
        '''
        For now just remove all hairpins that have no length.
        '''
        to_remove = []
        for d in self.defines:
            if d[0] == 'h' and len(self.defines[d]) == 0:
                to_remove += [d]

        for r in to_remove:
            self.remove_vertex(r)

    def from_stems_and_bulges(self, stems, bulges):
        '''
        Create the graph from the list of stems and bulges.

        @param stems: A list of tuples of two two-tuples, each containing the start
                      and end nucleotides of each strand of the stem.
        @param bulges: A list of tuples containing the starts and ends of the 
                       of the bulge regions.
        @return: Nothing, just make the bulgegraph
        '''
        for i in range(len(stems)):
            # one is added to each coordinate to make up for the fact that residues are 1-based
            ss1 = stems[i][0][0]+1
            ss2 = stems[i][0][1]+1
            se1 = stems[i][1][0]+1
            se2 = stems[i][1][1]+1

            self.defines['y%d' % (i)] = [min(ss1,se1), max(ss1,se1), 
                                         min(ss2,se2), max(ss2,se2)]
            self.weights['y%d' % (i)] = 1

        for i in range(len(bulges)):
            bulge = bulges[i]
            self.defines['b%d' % (i)] = sorted([bulge[0]+1, bulge[1]+1])
            self.weights['b%d' % (i)] = 1

        self.create_bulge_graph(stems, bulges)
        self.create_stem_graph(stems, len(bulges))
        self.collapse()
        self.relabel_nodes()
        self.remove_degenerate_nodes()
        self.sort_defines()
        
        # dissolve all stems which have a length of one
        if self.dissolve_length_one_stems:
            repeat=True
            while repeat:
                repeat = False
                for k in self.defines:
                    if k[0] == 's' and self.stem_length(k) == 1:
                        self.dissolve_stem(k)
                        repeat = True
                        break

    def from_dotbracket(self, dotbracket_str, dissolve_length_one_stems = False):
        '''
        Populate the BulgeGraph structure from a dotbracket representation.

        ie: ..((..))..

        :param dotbracket_str: A string containing the dotbracket representation
                               of the structure
        '''
        self.__init__()
        self.dissolve_length_one_stems = dissolve_length_one_stems
        (bulges, stems) = find_bulges_and_stems(dotbracket_str)

        self.dotbracket_str = dotbracket_str
        self.seq_length = len(dotbracket_str)

        self.from_stems_and_bulges(stems, bulges)

    def from_bpseq_str(self, bpseq_str, dissolve_length_one_stems = False):
        '''
        Create the graph from a string listing the base pairs.

        The string should be formatted like so:

            1 G 115
            2 A 0
            3 A 0
            4 U 0
            5 U 112
            6 G 111

        @param bpseq_str: The string, containing newline characters.
        @return: Nothing, but fill out this structure.
        '''
        self.__init__()
        lines = bpseq_str.split('\n')
        seq = ''

        line_iter = iter(lines)
        parts = next(line_iter).split(' ')
        #parts = line_iter.next().split(' ')

        stems = []
        bulges = []

        prev_from = (int(parts[0]))
        prev_to = (int(parts[2]))

        start_from = prev_from
        start_to = prev_to
        last_paired = prev_from

        seq = parts[1]

        self.dissolve_length_one_stems = dissolve_length_one_stems

        for line in line_iter:
            parts = line.split(' ')

            if len(parts) < 3:
                continue
            (from_bp, base, to_bp) = (int(parts[0]), parts[1], int(parts[2]))
            seq += base

            if abs(to_bp - prev_to) == 1 and prev_to != 0:
                # stem
                if ((prev_to - prev_from > 0 and to_bp - from_bp > 0) or
                    (prev_to - prev_from < 0 and to_bp - from_bp < 0)):
                    (prev_from, prev_to) = (from_bp, to_bp)
                    last_paired = from_bp
                    continue

            if to_bp == 0 and prev_to == 0:
                # bulge
                (prev_from, prev_to) = (from_bp, to_bp)
                continue
            else:
                if prev_to != 0:
                    new_stem = tuple(sorted([tuple(sorted([start_from - 1, start_to - 1])), 
                                tuple(sorted([prev_from - 1, prev_to - 1]))]))
                    if new_stem not in stems:
                        stems += [new_stem]

                    last_paired = from_bp
                    start_from = from_bp
                    start_to = to_bp
                else:
                    new_bulge = ((last_paired - 1, prev_from - 1))
                    bulges += [new_bulge]

                    start_from = from_bp
                    start_to = to_bp


            prev_from = from_bp
            prev_to = to_bp

        # Take care of the last element
        if prev_to != 0:
            new_stem = tuple(sorted([tuple(sorted([start_from - 1, start_to - 1])), 
                        tuple(sorted([prev_from - 1, prev_to - 1]))]))
            if new_stem not in stems:
                stems += [new_stem]
        if prev_to == 0:
            new_bulge = ((last_paired - 1, prev_from - 1))
            bulges += [new_bulge]

        self.seq = seq
        self.seq_length = len(seq)

        self.from_stems_and_bulges(stems, bulges)

    def sort_defines(self):
        '''
        Sort the defines of interior loops and stems so that the 5' region
        is always first.
        '''
        for k in self.defines.keys():
            d = self.defines[k] 

            if len(d) == 4:
                if d[0] > d[2]:
                    new_d = [d[2], d[3], d[0], d[1]]
                    self.defines[k] = new_d

    def to_dotbracket_string(self):
        '''
        Convert the BulgeGraph representation to a dot-bracket string
        and return it.

        @return: A dot-bracket representation of this BulgeGraph
        '''
        out = ['.' for i in range(self.seq_length)]
        for s in self.stem_iterator():
            for i in range(self.defines[s][0], self.defines[s][1]+1):
                out[i-1] = '('
            for i in range(self.defines[s][2], self.defines[s][3]+1):
                out[i-1] = ')'

        return "".join(out)

    def from_bg_file(self, bg_file):
        '''
        Load from a file containing a text-based representation
        of this BulgeGraph.

        @param bg_file: The filename.
        @return: No return value since the current structure is the one
                 being loaded.
        '''
        with open(bg_file, 'r') as f:
            bg_string = "".join(f.readlines())
            self.from_bg_string(bg_string)

    def from_bg_string(self, bg_str):
        '''
        Populate this BulgeGraph from the string created by the method
        to_bg_string.

        @param bg_str: The string representation of this BugleGraph.
        '''
        lines = bg_str.split('\n')
        for line in lines:
            line = line.strip()
            parts = line.split()
            if len(parts) == 0:
                #blank line
                continue
            if parts[0] == 'length':
                self.seq_length = int(parts[1])
            elif parts[0] == 'define':
                self.defines[parts[1]] = map(int, parts[2:])
            elif parts[0] == 'connect':
                for p in parts[2:]:
                    self.edges[parts[1]].add(p)
                    self.edges[p].add(parts[1])
            elif parts[0] == 'seq':
                self.seq = parts[1]
            elif parts[0] == 'seq_ids':
                self.seq_ids = map(ftum.parse_resid, parts[1:])
            elif parts[0] == 'name':
                self.name = parts[1].strip()

    def stem_iterator(self):
        '''
        Iterate over all of the stem elements.
        '''
        for d in self.defines.keys():
            if d[0] == 's':
                yield d

    def sorted_stem_iterator(self):
        '''
        Iterate over a list of the stems sorted by the lowest numbered
        nucleotide in each stem.
        '''
        stems = [d for d in self.defines if d[0] == 's']
        stems.sort(key=lambda s: self.defines[s][0])

        for s in stems:
            yield s

    def is_single_stranded(self, node):
        '''
        Does this node represent a single-stranded region?

        Single stranded regions are five-prime and three-prime unpaired
        regions, multiloops, and hairpins

        :param node: The name of the node
        :return: True if yes, False if no
        '''
        if node[0] == 'f' or node[0] == 't' or node[0] == 'm' or node[0] == 'h':
            return True
        else:
            return False

    def get_node_dimensions(self, node):
        '''
        Return the dimensions of a node.

        If the node is a stem, then the dimensions will be l where l is
        the length of the stem.

        Otherwise, see get_bulge_dimensions(node)

        :param node: The name of the node
        :return: A pair containing its dimensions
        '''
        if node[0] == 's':
            return (self.stem_length(node))
            '''
            return (self.defines[node][1] - self.defines[node][0] + 1,
                    self.defines[node][1] - self.defines[node][0] + 1)
            '''
        else:
            return self.get_bulge_dimensions(node)

    def adjacent_stem_pairs_iterator(self):
        '''
        Iterate over all pairs of stems which are separated by some element.

        This will always yield triples of the form (s1, e1, s2) where s1 and
        s2 are the stem identifiers and e1 denotes the element that separates
        them.
        '''
        for d in self.defines.keys():
            if len(self.edges[d]) == 2:
                edges = list(self.edges[d])

                if edges[0][0] == 's' and edges[1][0] == 's':
                    yield (edges[0], d, edges[1])

    def stem_bp_iterator(self, stem):
        '''
        Iterate over all the base pairs in the stem.
        '''
        d = self.defines[stem]
        stem_length = self.stem_length(stem)

        for i in range(stem_length):
            yield (d[0] + i, d[3] - i)

    def get_side_nucleotides(self, stem, side):
        '''
        Get the nucleotide numbers on the given side of
        them stem. Side 0 corresponds to the 5' end of the
        stem whereas as side 1 corresponds to the 3' side
        of the stem.

        @param stem: The name of the stem
        @param side: Either 0 or 1, indicating the 5' or 3' end of the stem
        @return: A tuple of the nucleotide numbers on the given side of
                 the stem.
        '''
        if side == 0:
            return (self.defines[stem][0], self.defines[stem][3])
        elif side == 1:
            return (self.defines[stem][1], self.defines[stem][2])

        raise Exception("Invalid side (%d) for the stem (%s)." % (stem, side))

    def get_any_sides(self, e1, e2):
        '''
        Get the side of e1 that e2 is on. The only difference from the get_sides
        method is the fact that e1 does not have to be a stem.

        0 indicates that e2 is on the side with lower numbered
        nucleotides and 1 indicates that e2 is on the side with
        greater nucleotide numbers.

        @param e1: The name of the first element.
        @param e2: The name of the second element.
        @return: A tuple indicating the side of e1 adjacent to e2 and the side of e2
                 adjacent to e1
        '''
        if e1[0] == 's':
            return self.get_sides(e1, e2)
        elif e2[0] == 's':
            return self.get_sides(e2, e1)[::-1]

        return None

    def get_sides(self, s1, b):
        '''
        Get the side of s1 that is next to b.

        s1e -> s1b -> b

        @param s1: The stem.
        @param b: The bulge.
        @return: A tuple indicating which side is the one next to the bulge
                 and which is away from the bulge.
        '''
        s1d = self.defines[s1]
        bd = self.defines[b]

        # if the bulge is a length 0, multiloop then use the adjacent
        # stem to determine its side
        if len(bd) == 0:
            edges = self.edges[b]

            for e in edges:
                if e != s1:
                    bd = self.defines[e]
                    break

        #print >>sys.stderr, "s1: %s b: %s" % (s1, b)

        for i in range(4):
            for k in range(len(bd)):
                if s1d[i] - bd[k] == 1:
                    if i == 0:
                        s1b = 0
                        break
                    if i == 2:
                        s1b = 1
                        break
                elif s1d[i] - bd[k] == -1:
                    if i == 1:
                        s1b = 1
                        break
                    if i == 3:
                        s1b = 0
                        break
        if s1b == 0:
            s1e = 1
        else:
            s1e = 0

        return (s1b, s1e)

    def get_sides_plus(self, s1, b):
        '''
        Get the side of s1 that is next to b.

        s1e -> s1b -> b

        @param s1: The stem.
        @param b: The bulge.
        @return: A tuple indicating the corner of the stem that connects
                 to the bulge as well as the corner of the bulge that connects
                 to the stem.
        '''
        s1d = self.defines[s1]
        bd = self.defines[b]

        #print >>sys.stderr, "s1: %s b: %s" % (s1, b)

        if len(bd) == 0:
            edges = self.edges[b]

            for e in edges:
                if e != s1:
                    bd = self.defines[e]
                    break

        #print >>sys.stderr, "s1: %s b: %s" % (s1, b)

        for k in range(len(bd)):
            # before the stem on the 5' strand
            if s1d[0] - bd[k] == 1:
                return (0, k)
            # after the stem on the 5' strand
            elif bd[k] - s1d[1] == 1:
                return (1, k)
            # before the stem on the 3' strand
            elif s1d[2] - bd[k] == 1:
                return (2, k)
            # after the stem on the 3' strand
            elif bd[k] - s1d[3] == 1:
                return (3, k)

        raise Exception("Faulty multiloop %s connecting %s"
                        % (" ".join(map(str, bd)),
                           " ".join(map(str, s1d))))

    def stem_side_vres_to_resn(self, stem, side, vres):
        '''
        Return the residue number given the stem name, the strand (side) it's on
        and the virtual residue number.
        '''
        d = self.defines[stem]

        if side == 0:
            return d[0] + vres
        else:
            return d[3] - vres

    def stem_iterator(self):
        '''
        Iterator over all of the stems in the structure.
        '''
        for d in self.defines.keys():
            if d[0] == 's':
                yield d
    
    def hloop_iterator(self):
        '''
        Iterator over all of the hairpin in the structure.
        '''
        for d in self.defines.keys():
            if d[0] == 'h':
                yield d

    def mloop_iterator(self):
        '''
        Iterator over all of the multiloops in the structure.
        '''
        for d in self.defines.keys():
            if d[0] == 'm':
                yield d

    def iloop_iterator(self):
        '''
        Iterator over all of the interior loops in the structure.
        '''
        for d in self.defines.keys():
            if d[0] == 'i':
                yield d

    def floop_iterator(self):
        '''
        Yield the name of the 5' prime unpaired region if it is
        present in the structure.
        '''
        if 'f1' in self.defines.keys():
            yield 'f1'

    def tloop_iterator(self):
        '''
        Yield the name of the 3' prime unpaired region if it is
        present in the structure.
        '''
        if 't1' in self.defines.keys():
            yield 't1'

    def pairing_partner(self, nucleotide_number):
        '''
        Return the base pairing partner of the nucleotide at position
        nucleotide_number. If this nucleotide is unpaired, return None.

        @param nucleotide_number: The position of the query nucleotide in the
                                  sequence.
        @return: The number of the nucleotide base paired with the one at
                 position nucleotide_number.
        '''
        for d in self.stem_iterator():
            for (r1, r2) in self.stem_bp_iterator(d):
                if r1 == nucleotide_number:
                    return r2
                elif r2 == nucleotide_number:
                    return r1
        return None

    def connections(self, bulge):
        '''
        Return the edges that connect to a bulge in a list form,
        sorted by lowest res number of the connection.
        '''
        connections = list(self.edges[bulge])
        connections.sort(key=lambda x: self.defines[x][0])

        return connections

    def get_define_seq_str(self, d, adjacent=False):
        '''
        Get an array containing the sequences for the given define.
        Non-stem sequences will contain the sequence without the overlapping
        stem residues that are part of the define.

        @param d: The define for which to get the sequences
        @return: An array containing the sequences corresponding to the defines
        '''
        define = self.defines[d]
        ranges = zip(*[iter(define)] * 2)
        c = self.connections(d)

        if d[0] == 'i':
            s1 = self.defines[c[0]]
            s2 = self.defines[c[1]]
            if adjacent:
                return [self.seq[s1[1]-1:s2[0]],
                        self.seq[s2[3]-1:s1[2]]]
            else:
                return [self.seq[s1[1]:s2[0]-1],
                        self.seq[s2[3]:s1[2]-1]]
        if d[0] == 'm':
            s1 = self.defines[c[0]]
            s2 = self.defines[c[1]]

            i1 = s1[self.get_sides_plus(c[0], d)[0]]
            i2 = s2[self.get_sides_plus(c[1], d)[0]]

            (i1, i2) = (min(i1, i2), max(i1, i2))

            if adjacent:
                return [self.seq[i1-1:i2]]
            else:
                return [self.seq[i1:i2-1]]
        else:
            seqs = [] 
            for r in ranges:
                if d[0] == 's':
                    seqs += [self.seq[r[0]-1:r[1]]]
                else:
                    if adjacent:
                        if r[0] > 1:
                            seqs += [self.seq[r[0]-2:r[1]+1]]
                        else:
                            seqs += [self.seq[r[0]-1:r[1]+1]]
                    else:
                        seqs += [self.seq[r[0]-1:r[1]]]
                
            return seqs

    def get_stem_direction(self, s1, s2):
        '''
        Return 0 if the lowest numbered residue in s1
        is lower than the lowest numbered residue in s2.
        '''
        if self.defines[s1][0] < self.defines[s2][0]:
            return 0
        return 1

    def get_multiloop_side(self, m):
        '''
        Find out which strand a multiloop is on. An example of a situation in
        which the loop can be on both sides can be seen in the three-stemmed
        structure below:

            (.().().)

        In this case, the first multiloop section comes off of the 5' strand of
        the first stem (the prior stem is always the one with a lower numbered
        first residue). The second multiloop section comess of the 3' strand of 
        the second stem and the third loop comes off the 3' strand of the third
        stem.
        '''
        c = self.connections(m)
        md = self.defines[m]
        s1 = self.defines[c[0]]
        s2 = self.defines[c[1]]

        p1 = self.get_sides_plus(c[0], m)
        p2 = self.get_sides_plus(c[1], m)

        return (p1[0], p2[0])

    def get_strand(self, multiloop):
        '''
        Get the strand on which this multiloop is located.

        @param multiloop: The name of the multiloop
        @return: 0 for being on the lower numbered strand and 1 for
                 being on the higher numbered strand.
        '''
        conn = self.connections(multiloop)
        t = self.connection_type(multiloop, conn)

        if abs(t) == 2:
            return 1
        elif abs(t) == 3:
            return 0
        else:
            return 0

        pass

    def get_bulge_dimensions(self, bulge):
        '''
        Return the dimensions of the bulge.

        If it is single stranded it will be (0, x). Otherwise it will be (x, y).

        @param bulge: The name of the bulge.
        @return: A pair containing its dimensions
        '''

        bd = self.defines[bulge]
        prev_stem = self.connections(bulge)[0]
        c = self.connections(bulge)

        if bulge[0] == 'i':
            # if this interior loop only has one unpaired region
            # then we have to find out if it's on the 5' strand or
            # the 3' strand
            # Example:
            # s1 1 3 
            #    23 25
            # s2 5 10 
            #    15 20
            s1 = self.defines[c[0]]
            s2 = self.defines[c[1]]

            dims = (s2[0] - s1[1] - 1, s1[2] - s2[3] - 1)

        if bulge[0] == 'm':
            # Multiloops are also pretty easy
            if len(bd) == 2:
                dims = (bd[1] - bd[0] + 1, 1000)
            else:
                dims = (0, 1000)
        if bulge[0] == 'f' or bulge[0] == 't':
            dims = (bd[1] - bd[0] + 1,-1)

        if bulge[0] == 'h':
            dims = (bd[1] - bd[0] + 1,-1)

        return dims

    def get_node_from_residue_num(self, base_num, seq_id=False):
        """
        Iterate over the defines and see which one encompasses this base.
        """
        for key in self.defines.keys():
            define = self.defines[key]

            for i in range(0, len(define), 2):
                a = [int(define[i]), int(define[i+1])]
                a.sort()

                if seq_id:
                    for i in range(a[0], a[1]+1):
                        if self.seq_ids[i-1][1] == base_num:
                            return key
                else:
                    if base_num >= a[0] and base_num <= a[1]:
                        return key

        raise Exception("Base number %d not found in the defines." % (base_num))


    def get_length(self, vertex):
        '''
        Get the minimum length of a vertex.

        If it's a stem, then the result is its length (in base pairs).

        If it's a bulge, then the length is the smaller of it's dimensions.

        @param vertex: The name of the vertex.
        '''
        if vertex[0] == 's':
            return abs(self.defines[vertex][1] - self.defines[vertex][0]) + 1
        else:
            if len(self.edges[vertex]) == 1:
                return self.defines[vertex][1] - self.defines[vertex][0] + 1
            else:
                dims = list(self.get_bulge_dimensions(vertex))
                dims.sort()

                if vertex[0] == 'i':
                    return sum(dims) / float(len(dims))

                else:
                    return min(dims)

    def get_flanking_region(self, bulge_name, side=0):
        '''
        If a bulge is flanked by stems, return the lowest residue number
        of the previous stem and the highest residue number of the next
        stem.

        @param bulge_name: The name of the bulge
        @param side: The side of the bulge (indicating the strand)
        '''
        c = self.connections(bulge_name)

        if bulge_name[0] == 'h':
            s1 = self.defines[c[0]]
            return (s1[0], s1[3])

        s1 = self.defines[c[0]]
        s2 = self.defines[c[1]]

        if bulge_name[0] == 'i':
            # interior loop
            if side == 0:
                return (s1[0], s2[1])
            else:
                return (s2[2], s1[3])

        elif bulge_name[0] == 'm':
            ss = self.get_multiloop_side(bulge_name)
            st = [s1, s2]

            ends = []

            # go through the two sides and stems and pick
            # the other end of the same strand
            for i, s in enumerate(ss):
                if s == 0:
                    ends += [st[i][1]]
                elif s == 1:
                    ends += [st[i][0]]
                elif s == 2:
                    ends += [st[i][3]]
                elif s == 3:
                    ends += [st[i][2]]
                else:
                    raise Exception("Weird multiloop sides: %s" %
                                    bulge_name)

            ends.sort()

            return tuple(ends)
            # multiloop

        return (m1, m2)


    def get_flanking_sequence(self, bulge_name, side=0):
        if len(self.seq) == 0:
            raise Exception("No sequence present in the bulge_graph: %s" % (self.name))

        (m1, m2) = self.get_flanking_region(bulge_name, side)

        return self.seq[m1-1:m2]

    def get_flanking_handles(self, bulge_name, side=0):
        '''
        Get the indices of the residues for fitting bulge regions.

        So if there is a loop like so (between residues 7 and 16):

        (((...))))
        7890123456
          ^   ^

        Then residues 9 and 13 will be used as the handles against which
        to align the fitted region.

        In the fitted region, the residues (2,6) will be the ones that will
        be aligned to the handles.

        @return: (orig_chain_res1, orig_chain_res1, flanking_res1, flanking_res2)
        '''
        def1 = self.defines[bulge_name]
        f1 = self.get_flanking_region(bulge_name, side)
        c = self.connections(bulge_name)

        if bulge_name[0] == 'h':
            s1 = self.defines[c[0]]
            ab = [s1[1], s1[2]]
            return (ab[0], ab[1], ab[0] - f1[0], ab[1] - f1[0])

        s1 = self.defines[c[0]]
        s2 = self.defines[c[1]]

        if bulge_name[0] == 'm':
            sides = self.get_multiloop_side(bulge_name)
            ab = [s1[sides[0]], s2[sides[1]]]
            ab.sort()

            return (ab[0], ab[1], ab[0] - f1[0], ab[1] - f1[0])

        if bulge_name[0] == 'i':
            if side == 0:
                ab = [s1[1], s2[0]]
            else:
                ab = [s2[3], s1[2]]

            return (ab[0], ab[1], ab[0] - f1[0], ab[1] - f1[0])

        # probably still have to include the 5' and 3' regions, but that
        # will come a little later
        return None
    
    def are_adjacent_stems(self, s1, s2, multiloops_count=True):
        '''
        Are two stems separated by only one element. If multiloops should not
        count as edges, then the appropriate parameter should be set.

        @param s1: The name of the first stem
        @param s2: The name of the second stem
        @param multiloops_count: Whether to count multiloops as an edge linking
                                 two stems
        '''
        for e in self.edges[s1]:
            if not multiloops_count and e[0] == 'm':
                continue
            if s2 in self.edges[e]:
                return True

        return False

    def random_subgraph(self, subgraph_length=None):
        '''
        Return a random subgraph of this graph.

        @return: A list containing a the nodes comprising a random subgraph
        '''
        if subgraph_length == None:
            subgraph_length = random.randint(1, len(self.defines.keys()))

        start_node = random.choice(self.defines.keys())
        curr_length = 0
        visited = set()
        next_nodes = [start_node]
        new_graph = []

        while curr_length < subgraph_length:
            curr_node = random.choice(next_nodes)
            if curr_node[0] == 'i' or curr_node[0] == 'm':
                # if it's an interior loop or a multiloop, then we have to
                # add the adjacent stems
                for e in self.edges[curr_node]:
                    if e in new_graph:
                        continue
                    visited.add(e)
                    new_graph += [e]
                    next_nodes += list(self.edges[e])
                    curr_length += 1

            visited.add(curr_node)
            next_nodes += list(self.edges[curr_node])
            next_nodes = [n for n in next_nodes if n not in visited]
            new_graph += [curr_node]
            curr_length += 1 #self.element_length(curr_node)

        return new_graph

    def same_stem_end(self, sd): 
        '''
        Return the index of the define that is on the same end of the
        stem as the index sd.

        @param sd: An index into a define.
        @return: The index pointing to the nucleotide on the other strand 
                 on the same side as the stem.
        '''
        if sd == 0: 
            return 3 
        elif sd == 1: 
            return 2 
        elif sd == 2: 
            return 1 
        else: 
            return 0 

    def extract_chain_id(self, chainres):
        '''
        Extract the chain identifier from a chain/residue
        identifier.

        @param chainres: A chain and residue identifier (i.e. 'A12', or '14')
        @return: A chain identifier.
        '''
        return ftum.parse_chain_base(chainres)[0]

    def extract_resnum(self, chainres):
        '''
        Extract the residue number from a chainres identifier.

        @param chainres: A chain and residue identifier (i.e. 'A12', or '14')
        @return: The residue number
        '''
        return ftum.parse_chain_base(chainres)[1]

    def get_resseqs(self, define, seq_ids=True):
        '''
        Return the pdb ids of the nucleotides in this define.

        @param define: The name of this element.
        @param: Return a tuple of two arrays containing the residue ids
                on each strand
        '''
        resnames = []
        ranges = zip(*[iter(self.defines[define])] * 2)

        for r in ranges:
            strand_resnames = []
            for x in range(r[0], r[1] + 1):
                if seq_ids:
                    strand_resnames += [self.seq_ids[x-1]]
                else:
                    strand_resnames += [x]

            resnames += [strand_resnames]

        return resnames

    def seqids_from_residue_map(self, residue_map):
        '''
        Create the list of seq_ids from the list of MC-Annotate identifiers in the
        residue map.
        '''
        self.seq_ids = []

        for r in residue_map:
            (from_chain, from_base) = ftum.parse_chain_base(r)

            self.seq_ids += [ftum.parse_resid(from_base)]
    
    def connected_stem_iterator(self):
        '''
        Iterate over all pairs of connected stems.
        '''
        for l in it.chain(self.mloop_iterator(), self.iloop_iterator()):
            edge_list = list(self.edges[l])
            yield (edge_list[0], l, edge_list[1])
    
    def get_mst(self):
        '''
        Create a minimum spanning tree from this BulgeGraph. This is useful
        for constructing a structure where each section of a multiloop is
        sampled independently and we want to introduce a break at the largest
        multiloop section.
        '''
        # keep track of all linked nodes
        sets = c.defaultdict(set)
        edges = sorted(it.chain(self.mloop_iterator(),
                                self.iloop_iterator()),
                       key=lambda x: min(self.get_node_dimensions(x)))

        mst = set(it.chain(self.stem_iterator(),
                           self.floop_iterator(),
                           self.tloop_iterator()))


        while len(edges) > 0:
            outside = False
            conn = edges.pop(0)
            neighbors = list(self.edges[conn])

            
            if len(set.intersection(sets[neighbors[0]], 
                                    sets[neighbors[1]])) == 0:
                # this edge joins two disconnected forests, so it is added
                # to the MST
                sets[neighbors[0]].add(neighbors[1])
                sets[neighbors[1]].add(neighbors[0])

                mst.add(conn)
                
        return mst

    def traverse_graph(self):
        '''
        Traverse the graph to get the angle types. The angle type depends on 
        which corners of the stem are connected by the multiloop or internal
        loop.
        '''
        if self.mst is None:
            self.mst = self.get_mst()

        build_order = []
        to_visit = [('s0', 'start')]
        visited = set(['s0'])
        while len(to_visit) > 0:
            (current, prev) = to_visit.pop(0)

            for e in self.edges[current]:
                if (e not in visited and e in self.mst):
                    # make sure the node hasn't been visited
                    # and is in the minimum spanning tree
                    to_visit.append((e, current))
                    visited.add(e)

            if current[0] != 's' and len(self.edges[current]) == 2:
                # multiloop or interior loop

                #overkill method of getting the stem that isn't
                #equal to prev
                next_stem = set.difference(self.edges[current],
                                           set([prev]))
                build_order += [(prev, current, list(next_stem)[0])]

        self.build_order = build_order
        return build_order

    def set_angle_types(self):
        '''
        Fill in the angle types based on the build order
        '''
        if self.build_order is None:
            self.traverse_graph()

        self.ang_types = dict()
        for (s1, b, s2) in self.build_order:
            self.ang_types[b] = self.connection_type(b, [s1,s2])

    def get_angle_type(self, bulge):
        '''
        Return what type of angle this bulge is, based on the way this
        would be built using a breadth-first traversal along the minimum
        spanning tree.
        '''
        if self.ang_types is None:
            self.set_angle_types()
        
        if bulge in self.ang_types:
            return self.ang_types[bulge]
        else:
            return None
        
def bg_from_subgraph(bg, sg):
    '''
    Create a BulgeGraph from a list containing the nodes
    to take from the original.

    WARNING: The sequence information is not copied
    '''
    nbg = BulgeGraph()
    nbg.seq_length = 0

    for d in sg:
        # copy the define
        nbg.defines[d] = bg.defines[d][::]

    # copy edges only if they connect elements which 
    # are also in the new structure
    for e in bg.edges.keys():
        for c in bg.edges[e]:
            if c in sg:
                nbg.edges[e].add(c)

    return nbg




#!/usr/bin/env python
import time
import math
import re
import pandas as pd
from pathlib import Path
import numpy as np
import subprocess
from Bio import SeqIO
import _pickle as cPickle
import os, sys
import matplotlib
matplotlib.use('agg')
from sklearn.preprocessing import OneHotEncoder
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import VarianceThreshold, chi2, f_classif
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from sklearn.metrics import make_scorer, confusion_matrix, matthews_corrcoef, roc_auc_score, roc_curve, auc
#from sklearn.model_selection import cross_validation
from sklearn.pipeline import Pipeline
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from scipy.spatial import cKDTree
import re
from Bio.Seq import Seq
# from Bio.Alphabet import generic_dna
import numpy as np
from mirge.libs.processSam import split_fasta_from_sam, combineSam, decorateSam, parse_refine_sam
from mirge.libs.generate_featureFiles import generate_featureFiles, get_precursors, renameStrFile
from mirge.libs.screen_precusor_candidates import screen_precusor_candidates
from mirge.libs.preprocess_featureFiles import preprocess_featureFiles, model_predict
from mirge.libs.write_novel_report import write_novel_report
from mirge.classes.exportHTML import FormatJS
# from sklearn.externals import joblib 
# /home/arun/.local/lib/python3.8/site-packages/sklearn/externals/joblib/__init__.py:15: FutureWarning: sklearn.externals.joblib is deprecated in 0.21 and will be removed in 0.23. Please import this functionality directly from joblib, which can be installed with: pip install joblib. If this warning is raised when loading pickled models, you may need to re-serialize those models with scikit-learn 0.21+.
# warnings.warn(msg, category=FutureWarning)


def convert2Fasta(pdUnmapped, infile, minLength, maxLength, countCutoff, outputdir2, species, speciesNameDic, base_names, rawReadCounts, filteredReadCounts):
    pdUnmapped = pdUnmapped.reset_index(level=['Sequence'])
    outf1 = open((Path(outputdir2)/"unmapped_mirna_raw.fa"), "w+")
    outf2 = open((Path(outputdir2)/"unmapped_mirna.fa"), "w+")
    cols1 = ["Sequence"] + base_names
    unmap = pd.DataFrame(pdUnmapped, columns= cols1)
    unmap['Sum'] =  unmap[base_names].sum(axis=1)
    finalColumns = ['Sequence','Sum']+base_names
    unmap = unmap.reindex(columns=finalColumns) # RAW dataframe
    unmap.index = unmap.index + 1 # Setting to start index from 1 instead of 0
    filtered_unmap = unmap[(unmap['Sequence'].str.len() >= minLength) & (unmap['Sequence'].str.len() <= maxLength) & (unmap['Sum'] >= countCutoff)] #Filtered Dataframe
    unmapList = unmap.reset_index().values.tolist()
    for index, row in enumerate(unmapList): 
        outf1.write(">mir"+str(row[0])+"_"+str(row[2])+"\n"+row[1]+"\n")
    filtered_unmapList = filtered_unmap.reset_index().values.tolist()
    for index, row in enumerate(filtered_unmapList):
        outf2.write(">mir"+str(row[0])+"_"+str(row[2])+"\n"+row[1]+"\n")
    for each_sample in base_names:
        current_fname_raw = "unmapped_mirna_"+each_sample+"_raw.fa"
        current_fname_filtered = "unmapped_mirna_"+each_sample+".fa"
        #outfn_raw = open((Path(outputdir2)/current_fname_raw), "w+") 
        outfn_filtered = open((Path(outputdir2)/current_fname_filtered), "w+")
        Sample_DF_raw = pd.DataFrame(unmap[unmap[each_sample] >= 1], columns=['Sequence', each_sample]) 
        rawReadCounts[each_sample] = Sample_DF_raw.shape[0]
        Sample_DF_filtered = pd.DataFrame(filtered_unmap[filtered_unmap[each_sample] >= countCutoff], columns=['Sequence', each_sample])
        filteredReadCounts[each_sample] = Sample_DF_filtered.shape[0]
        Sample_DF_rawList = Sample_DF_raw.reset_index().values.tolist() 
        Sample_DF_filteredList = Sample_DF_filtered.reset_index().values.tolist()
        #for index, row in enumerate(Sample_DF_rawList):
        #    outfn_raw.write(">mir"+str(row[0])+"_"+str(row[2])+"\n"+row[1]+"\n")
        #outfn_raw.close()
        for index, row in enumerate(Sample_DF_filteredList):
            outfn_filtered.write(">mir"+str(row[0])+"_"+str(row[2])+"\n"+row[1]+"\n")
        outfn_filtered.close()
    outf1.close()
    outf2.close()
    
def cluster_basedon_location(outfile2, overlapLenThresholdTmp,files, outputdir2, cluster_file):
    # The arguments of this function are: unmapped_mirna_vs_genome_sorted.sam(Or mapped_mirna_vs_genome_sorted.sam) overlapLenThreshold(12, this value must be less than  16)
    overlapLenThreshold = int(overlapLenThresholdTmp)
    inf = open(outfile2, 'r')
    chrList = []
    chrSeqDic = {}
    for line in inf:
        if line[0] != '@':
            contentTemp = line.strip().split('\t')
            chr = contentTemp[2]
            if 'chr' in chr:
                if chr not in chrList:
                    chrList.append(chr)
                    chrSeqDic.update({chr:[[],[]]})
                    #in chrSeqDic[chr], the fist list store the information about "+" forward strand match. #and the second one store the information about "-" reverse strand match
                flag = contentTemp[1]
                startPos = int(contentTemp[3])
                seq = contentTemp[9]
                endPos = startPos+len(seq)-1
                mirID = contentTemp[0]
                if flag == "0" and len(chrSeqDic[chr][0]) == 0:
                    chrSeqDic[chr][0].append([startPos, endPos, seq, [mirID]])
                elif flag == "16" and len(chrSeqDic[chr][1]) == 0:
                    chrSeqDic[chr][1].append([startPos, endPos, seq, [mirID]])
                else:
                    if flag == "0":
                        temp = chrSeqDic[chr][0][-1]
                        if startPos >= temp[0] and startPos <= temp[1] and temp[1]-startPos+1 >= overlapLenThreshold:
                            mirIDList = temp[3]
                            mirIDList.append(mirID)
                            startPosNew = temp[0]
                            if endPos <= temp[1]:
                                endPosNew = temp[1]
                                seqNew = temp[2]
                            else:
                                endPosNew = endPos
                                seqNew = temp[2] + seq[temp[1]-startPos+1:]
                            chrSeqDic[chr][0][-1] = [startPosNew, endPosNew, seqNew, mirIDList]
                        else:
                            chrSeqDic[chr][0].append([startPos, endPos, seq,[mirID]])
                    elif flag == "16":
                        temp = chrSeqDic[chr][1][-1]
                        if startPos >= temp[0] and startPos <= temp[1]and temp[1]-startPos+1 >= overlapLenThreshold:
                            mirIDList = temp[3]
                            mirIDList.append(mirID)
                            startPosNew = temp[0]
                            if endPos <= temp[1]:
                                endPosNew = temp[1]
                                seqNew = temp[2]
                            else:
                                endPosNew = endPos
                                seqNew = temp[2] + seq[temp[1]-startPos+1:]
                            chrSeqDic[chr][1][-1] = [startPosNew, endPosNew, seqNew, mirIDList]
                    else:
                        chrSeqDic[chr][1].append([startPos, endPos, seq,[mirID]])
    inf.close()
    with open(cluster_file,'w') as outf:
        i = 1
        outf.write('miRClusterID\tChr\tStrand\tStart\tEnd\tSequence\tSequenceLenght\tCoutOfReads\tCountOfMembers\tMembers\n')
        for chr in chrList:
            for content in chrSeqDic[chr][0]:
                readCount = 0
                for item in content[3]:
                    readCount = readCount + int(item.split('_')[1])
                outf.write('\t'.join([files+':'+'miRCluster_'+str(i)+'_'+str(len(content[2])),chr, "+", str(content[0]), str(content[1]), content[2],str(len(content[2])), str(readCount), str(len(content[3])),','.join(content[3])])+ '\n')
                i = i + 1
            for content in chrSeqDic[chr][1]:
                readCount = 0
                for item in content[3]:
                    readCount = readCount + int(item.split('_')[1])
                outf.write('\t'.join([files+':'+'miRCluster_'+str(i)+'_'+str(len(content[2])),chr, "-", str(content[0]), str(content[1]), content[2],str(len(content[2])), str(readCount), str(len(content[3])),','.join(content[3])])+ '\n')
                i = i + 1
    return i

def preTrimClusteredSeq(CoordinateDic, cluster_file, files, clusterSeqLenCutoff, clusterTrimedFile, clusterTrimedFile_orig_FASTA):
    inf1 = open(cluster_file, 'r')
    outf = open(clusterTrimedFile, 'w')
    outFA = open(clusterTrimedFile_orig_FASTA, 'w')
    line = inf1.readline()
    outf.write("\t".join(['miRClusterID\tOriginalSeq\tFlag\trepetitiveElementName']+line.split('\t')[1:]))
    line = inf1.readline()    
    trimmed_count=1
    while line != "":
        # flag = "0" means the seq will be abandoned. flag = "1" means the seq will be retained.
        flag = "0"
        repetitiveElementName = '*'
        content = line.strip().split('\t')
        # remove representative seqs with ength larger than 30
        if content[2] == '-':
            seq = Seq(content[5])
            orginalSeq = str(seq.reverse_complement())
        else:
            orginalSeq = content[5]
        if int(content[6]) <= int(clusterSeqLenCutoff):
            chr = content[1]
            start = int(content[3])
            end = int(content[4])
            # remove representative seqs with "AAAAAA" in the 3' and "TTTTTT" in the 5'
            if (re.search('A{6,}$', orginalSeq) is None) and (re.search('^T{6,}', orginalSeq) is None):
                if chr in CoordinateDic.keys():
                    dist, idx =CoordinateDic[chr][0][0].query([(start, 0)], 2)
                    if len(idx) == 1 and np.isfinite(dist[0][0]) and np.isfinite(dist[0][1]):
                        candidate1 = CoordinateDic[chr][1][idx[0][0]]
                        candidate2 = CoordinateDic[chr][1][idx[0][1]]
                        list1 = range(candidate1[0],candidate1[1]+1)
                        list2 = range(candidate2[0],candidate2[1]+1)
                        listSeq = range(start, end+1)
                        if len(set(list1) & set(listSeq)) == 0 and len(set(list2) & set(listSeq)) == 0:
                            flag = "1"
                        else:
                            if len(set(list1) & set(listSeq)) > 0:
                                repetitiveElementName = candidate1[2]
                            else:
                                repetitiveElementName = candidate2[2]
                    else:
                        candidate1 = CoordinateDic[chr][1][idx[0][0]]
                        list1 = range(candidate1[0],candidate1[1]+1)
                        listSeq = range(start, end+1)
                        if len(set(list1) & set(listSeq)) == 0:
                            flag = "1"
                        else:
                            repetitiveElementName = candidate1[2]
                else:
                    flag = "1"
        outf.write('\t'.join([content[0], orginalSeq, flag, repetitiveElementName]+content[1:])+'\n')
        if flag == "1":
            trimmed_count+=1
            outFA.write('>'+content[0] + ":" + chr + ":" + str(start)+"_"+str(end) + content[2]+"\n"+orginalSeq+"\n")
        line = inf1.readline()
    inf1.close()
    outf.close()
    outFA.close()
    return trimmed_count


def predict_nmir(args, workDir, ref_db, base_names, pdUnmapped):
    htmlJS = FormatJS(workDir)
    htmlJS.openNovelmiRJSData()
    predict_start_time = time.perf_counter()
    runlogFile = Path(workDir)/"run.log"
    outlog = open(str(runlogFile),"a+")
    if not args.quiet:
        print('\nPerforming prediction of novel miRNAs...')
        print('Start to predict')
    outlog.write('\nPerforming prediction of novel miRNAs...')
    outlog.write('Start to predict')
    samtoolsBinary = Path(args.samtools_path)/"samtools " if args.samtools_path else "samtools " 
    rnafoldBinary = Path(args.RNAfold_path)/"RNAfold " if args.RNAfold_path else "RNAfold "
    species = args.organism_name
    genomRepeats = species+"_genome_repeats.pckl"
    genomFasta = species+"_genome.pckl"
    genom_miRDB = species+"_mature_"+ref_db+".fa"
    genome_repeats = Path(args.libraries_path)/args.organism_name/"annotation.Libs"/genomRepeats
    genome_fa = Path(args.libraries_path)/args.organism_name/"fasta.Libs"/genomFasta
    mature_miRNA_fa = Path(args.libraries_path)/args.organism_name/"fasta.Libs"/genom_miRDB
    nameAbbrNameDic ={'human':'hsa', 'zebrafish':'dre', 'mouse':'mmu', 'rat':'rno', 'fruitfly':'dme', 'nematode':'cel', 'hamster':'mau'}
    if species in nameAbbrNameDic.keys():
        abbrName = nameAbbrNameDic[species]
        speciesNameDic = {species:abbrName}
    else:
        speciesNameDic = {species:"CUS"}

    if args.organism_name  not in ['human', 'mouse']:
        speciesType = 'others'
    else:
        speciesType = args.organism_name
    #wideSampleListFile = args.wideSampleListFile
    minLength = int(args.minLength)
    maxLength = int(args.maxLength)
    countCutoff = int(args.minReadCounts)
    mapping_loc = int(args.maxMappingLoci)
    seedLength = int(args.seedLength)
    # Here, the optimal overlapLenCutoff should be set to be 14, based on the trial of 8, 10 , 12
    overlapLenCutoff = int(args.overlapLenCutoff)
    clusterSeqLenCutoff = int(args.clusterLength)
    genome_idx_file = species+'_genome.1.ebwt'
    genome_idx = species+'_genome'
    indexPath = Path(args.libraries_path)/args.organism_name/'index.Libs'/genome_idx_file
    genome_index = Path(args.libraries_path)/args.organism_name/'index.Libs'/genome_idx
    if not Path(indexPath).exists():
        print(f'ERROR: The bowtie index file of {species}_genome.*.ebwt is not located at {indexPath}, please check it.')
        outlog.write(f'ERROR: The bowtie index file of {species}_genome.*.ebwt is not located at {indexPath}, please check it.')
        exit()
    try:
        with open(genome_repeats, 'rb') as f:
            repEleChrCoordinateDic = cPickle.load(f)
    except IOError:
        repEleChrCoordinateDic = {}
    with open(genome_fa, 'rb') as f:
        chrSeqDic = cPickle.load(f)
    chrSeqLenDic = {}
    for key in chrSeqDic.keys():
        chrSeqLenDic.update({key:len(chrSeqDic[key])})
    #Load genome miRNA fasta file
    exactmiRNASeqDic = {}
    for record in SeqIO.parse(mature_miRNA_fa, "fasta"):
        exactmiRNASeqDic.update({record.id:str(record.seq)})
    #Load the genome miRNA coordinate file
    mc_name = args.organism_name+'_'+ref_db+'.gff3'
    miRNA_coordinate = Path(args.libraries_path)/args.organism_name/"annotation.Libs"/mc_name
    miRNAchrCoordivateDic = {}
    strandLabel=""
    with open(miRNA_coordinate,"r") as inf1:
        for line1 in inf1:
            if line1[0] != "#":
                content = line1.strip().split("\t")
                if content[2] == "miRNA":
                    chr = content[0]
                    if ref_db == 'miRBase':
                        miRNAName = content[-1].split(";")[2].split("=")[-1]
                    else:
                        miRNAName = content[-1].split(";")[0].split("=")[-1]
                        startPos = int(content[3])
                        endPos = int(content[4])
                        strandLabel = content[6]
                    if chr not in miRNAchrCoordivateDic.keys(): 
                        #in miRNAchrCoordivateDic[chr], the fist list store the information about "+" forward strand match.
                        #and the second one store the information about "-" reverse strand match
                        miRNAchrCoordivateDic.update({chr:[[],[]]})
                    else:
                        pass
                    if strandLabel == "+":
                        miRNAchrCoordivateDic[chr][0].append((startPos, endPos, miRNAName))
                    elif strandLabel == "-":
                        miRNAchrCoordivateDic[chr][1].append((startPos, endPos, miRNAName))
                    else:
                        pass
                else:
                    pass
            else:
                pass
    ### Arguments preprocession is done!
    # Perform novel miRNA detection if selecting predict mode.

    # To modify the variable infile
    #infile = args.infile
    #Load the output file 'maaped.csv' and 'unmapped.csv' for downstream analysis representatively.
    outputdir2 = Path(workDir)/"unmapped_tmp"
    os.mkdir(outputdir2)

    bwtCmdTmp = str(Path(args.bowtie_path)/"bowtie ") if args.bowtie_path else "bowtie "
    if args.bowtieVersion == "False": # That is if version is v1.3.0
        bwtCmdTmp += " -x "
    bwtBuildCmdTmp = Path(args.bowtie_path)/"bowtie-build " if args.bowtie_path else "bowtie-build " 
    samtoolsCmdTmp = Path(args.samtools_path)/"samtools " if args.samtools_path else "samtools "
    rnafoldCmdTmp = Path(args.RNAfold_path)/"RNAfold " if args.RNAfold_path else "RNAfold "
    infile = "unmapped.log"
    logFile = Path((Path(outputdir2).resolve().parents[0]))/infile
    #logFile = Path(outputdir2)/infile
    rawReadCounts={}
    filteredReadCounts={}
    with open(logFile, 'w') as outfLog:
        time1 = time.perf_counter()
        outfLog.write(f'Reading and trimming unmapped reads from Pandas Library\n')
        convert2Fasta(pdUnmapped, infile, minLength, maxLength, countCutoff, outputdir2, species, speciesNameDic, base_names,rawReadCounts,filteredReadCounts)
        #print(rawReadCounts, filteredReadCounts)
        #time.sleep(.555)
        time2 = time.perf_counter()
        outfLog.write('Execution time for reading and trimming unmapped Pandas Library took: %.4fs\n'%(time2-time1))
        outfLog.write('********************\n')
        outfLog.flush()
        for files in base_names:
            errorTrue = 0
            outfLog.write(f'Processing {files}\n')
            outfLog.write(f'**There are {str(rawReadCounts[files])} collapsed reads in the raw fasta file\n')
            outfLog.write(f'**After filtering, there are {str(filteredReadCounts[files])} reads left\n')
            outfLog.write('Mapping reads to %s genome\n'%(species))
            time3 = time.perf_counter()
            current_fname_filtered = "unmapped_mirna_"+ files +".fa"
            fileNameTemp = Path(outputdir2)/current_fname_filtered
            outfile1 = Path(outputdir2)/("unmapped_mirna_"+ files +"_vs_genome.sam")
            bwtExec = str(bwtCmdTmp) + str(genome_index) + " " + str(fileNameTemp) + " -f -n 0 --best -a --threads " + str(args.threads) + " -m " + str(mapping_loc) + " -l "+ str(seedLength) + " -S " + str(outfile1)
            bowtie = subprocess.run(str(bwtExec), shell=True, check=True, stdout=subprocess.PIPE, text=True, stderr=subprocess.PIPE, universal_newlines=True)
            # SORT SAM FILE
            outfile2 = Path(outputdir2)/("unmapped_mirna_"+ files +"_vs_genome_sorted.sam") 
            samyExec = str(samtoolsCmdTmp) + " sort --threads "+ str(args.threads) + " -O sam -T sample.sort -o " + str(outfile2) + " " + str(outfile1)
            samysort = subprocess.run(str(samyExec), shell=True, check=True, stdout=subprocess.PIPE, text=True, stderr=subprocess.PIPE, universal_newlines=True)
            time4 = time.perf_counter()
            outfLog.write('Mapping reads to humna genome time: %.4fs\n'%(time4-time3))
            outfLog.write('Clustering the reads based on the coordinate in the genome\n')
            cluster_file = Path(outputdir2)/(files+"_clusters.tsv")
            clusters_log = cluster_basedon_location(outfile2, overlapLenCutoff, files, outputdir2, cluster_file)
            time5 = time.perf_counter()
            outfLog.write('Clustering the reads based on the coordinate in the genome time: %.4fs\n'%(time5-time4))
            outfLog.write('Trimming the clustered sequences (preliminary filtering)\n')
            outfLog.write(f'**Before preliminary trimming, there are {clusters_log} clustered sequences\n')
            clusterTrimedFile = Path(outputdir2)/(files+"clusters_trimmed.tsv")
            clusterTrimedFile_orig_FASTA = Path(outputdir2)/(files+"_clusters_trimmed_orig.fa")
            trimedClusters_log = preTrimClusteredSeq(repEleChrCoordinateDic, cluster_file, files, clusterSeqLenCutoff, clusterTrimedFile, clusterTrimedFile_orig_FASTA)
            time6 = time.perf_counter()
            outfLog.write('Trimming the clustered sequences time: %.fs\n'%(time6-time5))
            outfLog.write(f'**After preliminary trimming, there are {trimedClusters_log} the clustered sequences\n')
            outfLog.write('Generating the orignial seqeneces of the trimmed clustered sequences\n')
            outfLog.write('Building bowtie index file and mapping reads to the clustered seqneces\n')
            time7 = time.perf_counter()
            #os.system('%s -f %s_clusters_trimmed_orig.fa %s_representative_seq >> %s 2>&1'%(bwtBuildCmdTmp, outfile2, outfile3, outLogTemp))
            outfile3 = Path(outputdir2)/(files+"_representative_seq")
            bwtBuildExec = str(bwtBuildCmdTmp) +" -f "+ str(clusterTrimedFile_orig_FASTA) + " " + str(outfile3) + " --threads " + str(args.threads) 
            try:
                bowtie = subprocess.run(str(bwtBuildExec), shell=True, check=True, stdout=subprocess.PIPE, text=True, stderr=subprocess.PIPE, universal_newlines=True)
            except subprocess.CalledProcessError:
                errorTrue = 1
                pass
            # Align the total reads to the cluster sequences with exact match. '--norc' only considers matching to the forwad reference strand.
            outfile4 = Path(outputdir2)/(files+"_tmp1.sam")
            bwt2Exec = str(bwtCmdTmp) + str(outfile3) + " " + str(fileNameTemp) + " -f -n 0 --best -a --norc --threads " + str(args.threads) + " -m " + str(mapping_loc) + " -l "+ str(seedLength) + " -S " + str(outfile4)
            try:
                bowtie = subprocess.run(str(bwt2Exec), shell=True, check=True, stdout=subprocess.PIPE, text=True, stderr=subprocess.PIPE, universal_newlines=True)
            except subprocess.CalledProcessError:
                errorTrue = 1
                pass
            #Generate the reads that can't align to the cluster reads.
            try:
                imperfect_FASTA=Path(outputdir2)/(files+"_imperfectMath2Cluster.fa")
                split_fasta_from_sam(outfile4, fileNameTemp, str(imperfect_FASTA))
            except IOError:
                errorTrue = 1
                outlog.write(f'No cluster sequences are generated and prediction is aborted for {files}.\n')
                #os.system('rm -r %s'%(outputdir2))
                #sys.exit(1)
            # Align the rest part of reads to the cluster sequences with up to 1 mismatches with special 5 vs 3 prime considerations.
            # The alignments in the best stratum are those having the least number of mismatches.
            # '--norc' only considers matching to the forwad reference strand.
            if errorTrue != 1:
                outfile4_tmp2 = Path(outputdir2)/(files+"_tmp2.sam")
                bwt3Exec = str(bwtCmdTmp) + str(outfile3) + " " + str(imperfect_FASTA) + " -f -n 1 -l 15 -5 1 -3 3 --best --strata -a --norc --threads " + str(args.threads) + " -S " + str(outfile4_tmp2)
                bowtie = subprocess.run(str(bwt3Exec), shell=True, check=True, stdout=subprocess.PIPE, text=True, stderr=subprocess.PIPE, universal_newlines=True)
                # Combine the aligned result of the two type of reads: perfect matched reads and imperfect matched reads.
                combined_Sam = Path(outputdir2)/(files+".sam")  
                combineSam(str(outfile4), str(outfile4_tmp2), str(combined_Sam))
                time8 = time.perf_counter()
                outfLog.write('Building bowtie index file and mapping reads to the clustered seqneces time: %.1fs\n'%(time8-time7))
                outfLog.write('Decorating bowtie output sam files\n')
                outfile_modifiedSam = Path(outputdir2)/(files+"_modified.sam")
                outfile_modSam_RepSeq = Path(outputdir2)/(files+"_RepSeq_modified.sam")
                decorateSam(str(combined_Sam), str(fileNameTemp), str(outfile_modifiedSam), str(clusterTrimedFile_orig_FASTA)) #Decorate SAM handles two different inputs and produce two outputs!!!
                decorateSam(str(outfile2), str(fileNameTemp), str(outfile_modSam_RepSeq))
                outfLog.write('Decorating is done\n')
                outfileSelectTSV = Path(outputdir2)/(files+"_selected.tsv")
                outfileRevKeptTSV = Path(outputdir2)/(files+"_selected_reverseKept.tsv")
                parse_refine_sam(str(outfile_modifiedSam), str(outfileSelectTSV), str(outfileRevKeptTSV))
                #os.system('rm %s.bam %s.sam %s.bai %s.bam'%(outfile1, outfile1, outfile2, outfile2))
                os.system('sort -k6,6 -k1,1 %s > %s'%(str(outfileSelectTSV), str(Path(outputdir2)/(files+"_modified_selected_sorted.tsv"))))
                os.system('sort -k6,6 -k1,1 %s > %s'%(str(outfileRevKeptTSV), str(Path(outputdir2)/(files+"_modified_selected_reverseKept_sorted.tsv"))))
                # Trimming the clustered seuences based on the alligned results of all the reads (secondary filtering)
                #generate_featureFiles(outfile4+'_modified_selected_sorted.tsv', chrSeqDic, chrSeqLenDic, miRNAchrCoordivateDic, exactmiRNASeqDic)
                generate_featureFiles(str(Path(outputdir2)), files, chrSeqDic, chrSeqLenDic, miRNAchrCoordivateDic, exactmiRNASeqDic)
                outfLog.write('feature file is generated\n')
                outfLog.write('Generating the precusors based on the clustered sequences\n')
                time9 = time.perf_counter()
                precursor_count = get_precursors(str(Path(outputdir2)), files, chrSeqDic)
                time10 = time.perf_counter()
                outfLog.write('Generating the precusors based on the clustered sequences time: %.1fs\n'%(time10-time9))
                outfLog.write(f'**There are {str(precursor_count)} precusors generated\n')
                outfLog.write('Folding precursors with RNAfold\n')
                time11 = time.perf_counter()
                infile_pre = str(Path(outputdir2)/(files+"_precursor.fa"))
                outfile_str = str(Path(outputdir2)/(files+"_precursor_tmp.str"))
                rnafld_exec = str(rnafoldCmdTmp) + " " + str(infile_pre) + " --noPS --noLP > " + str(outfile_str)
                rnafldRun = subprocess.run(str(rnafld_exec), shell=True, check=True, stdout=subprocess.PIPE, text=True, stderr=subprocess.PIPE, universal_newlines=True)
                strFileOut= str(Path(outputdir2)/(files+"_precursor.str"))
                renameStrFile(infile_pre, outfile_str, strFileOut)
                time12 = time.perf_counter()
                outfLog.write('Folding precursors with RNAfold time: %.1fs\n'%(time12-time11))
                screen_precusor_candidates(str(Path(outputdir2)), files, str(Path(outputdir2)/(files+"_features.tsv")), strFileOut, str(rnafoldCmdTmp))
                outfLog.write('********************\n')
                outfLog.flush()
                modelDirTmp = Path(__file__).resolve().parents[1]
                modelDir = str(Path(modelDirTmp)/'models')
                fileToPredict = Path(outputdir2)/(files+'_updated_stableClusterSeq_15.tsv')
                preprocess_featureFiles(str(Path(outputdir2)), files, fileToPredict, str(Path(modelDir)/'total_features_namelist.txt'))
                speciesType = args.organism_name
                if args.organism_name == "human" or args.organism_name == "mouse":
                    mf = str(Path(modelDir)/(speciesType+'_svc_model.pkl'))
                else:
                    mf = str(Path(modelDir)/"others_svc_model.pkl")
                model_predict(str(outputdir2), files, mf)
                #model_predict(str(outputdir2), files, str(Path(modelDir)/(speciesType+'_svc_model.pkl')))
                novelmiRNALListFile = str(Path(outputdir2)/(files+'_novel_miRNAs_miRge2.0.csv'))
                featureFile = fileToPredict
                clusterFile = str(Path((outputdir2)/(files+'_cluster.txt')))
                write_novel_report(novelmiRNALListFile, featureFile, clusterFile, str(rnafoldCmdTmp), str(Path(outputdir2)), files)
    if errorTrue ==1: 
        print(f'No cluster sequences are generated and prediction is aborted.')
    predict_end_time = time.perf_counter()
    os.system('rm -r %s'%(outputdir2))
    htmlJS.closeNovelmiRJSData()
    if not args.quiet:
        print('Prediction of novel miRNAs Completed (%.2f sec)'%(predict_end_time - predict_start_time))
    outlog.write('Prediction of novel miRNAs Completed (%.2f sec)'%(predict_end_time - predict_start_time))
    outlog.close()



import os
from pathlib import Path
import subprocess
from reportlab.lib import colors
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import A4
from math import sqrt, asin, cos, sin
from Bio.Seq import Seq
# from Bio.Alphabet import generic_dna
import re
from mirge.classes.exportHTML import FormatJS

def Shifting(xcDic, ycDic):
    minx = min([xcDic[key] for key in xcDic.keys()])
    maxx = max([xcDic[key] for key in xcDic.keys()])
    miny = min([ycDic[key] for key in ycDic.keys()])
    maxy = max([ycDic[key] for key in ycDic.keys()])

    ## now place sec-structure in upper right corner
    shiftx = abs(minx)+10;
    shifty = abs(miny)+10;

    #shift everything to printable area;
    if minx < 0:
        for i in range(len(xcDic)):
            xcDic[i] = xcDic[i] + shiftx 
    if miny < 0:
        for i in range(len(ycDic)):
            ycDic[i] = ycDic[i] + shifty
    if maxx > 595.27:
        for i in range(len(xcDic)):
            xcDic[i] = xcDic[i] - (minx +10)
    if maxy > 841.89:
        for i in range(len(ycDic)):
            ycDic[i] = ycDic[i] - (miny + 10)

def centroidOfRectangle(minx, maxx, miny, maxy):
    return ((minx+maxx)/2, (miny+maxy)/2)

def lenAndWidOfRectangle(minx, maxx, miny, maxy):
    return (maxx-minx, maxy-miny)

def chunkListNew(number):
    outList = []
    if number <= 35:
        outList.append(range(number))
    else:
        outList.append(range(35))
        newList = range(35, number)
        newListTmp = [newList[i:i+67] for i in range(0, len(newList), 67)]
        for itmTmp in newListTmp:
            outList.append(itmTmp)
    return outList


def creatPDF(sampleName, novelmiRNANameNew, probability, chr, startPos, endPos, strand, armType, readCountSumMatureMiRNA, totalReadCountSum, matureMiRNAPrecusorSeq, matureMiRNAPrecusorStr, matureMiRNASeq, passengerMiRNASeq, psFile, outFile, matureReadContentlist, passengerReadContentlist):
    matureMiRNALocationList = range(matureMiRNAPrecusorSeq.index(matureMiRNASeq), matureMiRNAPrecusorSeq.index(matureMiRNASeq)+len(matureMiRNASeq))
    if passengerMiRNASeq != 'None':
        try:
            starMiRNALocationList = range(matureMiRNAPrecusorSeq.index(passengerMiRNASeq), matureMiRNAPrecusorSeq.index(passengerMiRNASeq)+len(passengerMiRNASeq))
        except ValueError:
            starMiRNALocationList = []
    else:
        starMiRNALocationList = []

    c = canvas.Canvas(outFile, pagesize=A4)
    xposshift = 15
    downy = 20
    # Write the titile
    c.setFont('Courier', 15)
    c.drawString(xposshift+180, downy+750, 'miRge 3.0 Novel miRNA')

    # Input the sequence information
    c.setFont('Courier', 7)
    c.drawString(xposshift+15, downy+700, 'Name: ')
    c.drawString(xposshift+148, downy+700, novelmiRNANameNew)
    c.drawString(xposshift+15, downy+690, 'Sample name: ')
    c.drawString(xposshift+148, downy+690, sampleName)
    c.drawString(xposshift+15, downy+680, 'Probability: ')
    c.drawString(xposshift+148, downy+680, '%.5f'%(float(probability)))
    c.drawString(xposshift+15, downy+670, 'Chr(GRCh38): ')
    c.drawString(xposshift+148, downy+670, chr)
    c.drawString(xposshift+15, downy+660, 'Start position of mature miRNA: ')
    c.drawString(xposshift+148, downy+660, str(startPos))
    c.drawString(xposshift+15, downy+650, 'End position of mature miRNA: ')
    c.drawString(xposshift+148, downy+650, str(endPos))
    c.drawString(xposshift+15, downy+640, 'Strand: ')
    c.drawString(xposshift+148, downy+640, strand)
    c.drawString(xposshift+15, downy+630, 'Total read count: ')
    c.drawString(xposshift+148, downy+630, str(totalReadCountSum))
    c.drawString(xposshift+15, downy+620, 'Mature miRNA read count: ')
    c.drawString(xposshift+148, downy+620, str(readCountSumMatureMiRNA))
    c.drawString(xposshift+15, downy+610, 'Passenger miRNA read count: ')
    c.drawString(xposshift+148, downy+610, str(int(totalReadCountSum)-int(readCountSumMatureMiRNA)))
    
    # Draw secondary structures.
    xcDic = {}
    ycDic = {}
    bpDic = {}
    twisted = 0 ## if mature sequence comes after loop ('arm3') twisted is 1
    if armType == 'arm3':
        twisted = 1

    mb = matureMiRNAPrecusorSeq.index(matureMiRNASeq)
    me = mb + len(matureMiRNASeq) - 1     # mature end coordinate
    centering_x = 0
    with open(psFile, 'r') as infTmp:
        line = infTmp.readline()
        flag = 0
        counter = 0
        while line != '':
            if line.strip().startswith('/sequence'):
                line = infTmp.readline()
                sequence = line.strip()[:-1]
            if line.strip() == '/coor [':
                flag = 1
                line = infTmp.readline()
            if line.strip() == '] def':
                flag = 0
            if line.strip() == '/pairs [':
                flag = 2
                line = infTmp.readline()
            if flag == 1:
                xc = float(line.strip()[1:-1].split(' ')[0])
                yc = float(line.strip()[1:-1].split(' ')[1])
                xcDic.update({counter:xc})
                ycDic.update({counter:yc})
                counter = counter + 1
            if flag == 2:
                if twisted == 1:
                    bpo2r = int(line.strip()[1:-1].split(' ')[0])
                    bpo1r = int(line.strip()[1:-1].split(' ')[1])
                else:
                    bpo2r = int(line.strip()[1:-1].split(' ')[1])
                    bpo1r = int(line.strip()[1:-1].split(' ')[0])
                bpDic.update({bpo1r-1:bpo2r-1})
            line = infTmp.readline()
    centering_x = matureMiRNALocationList[len(matureMiRNALocationList)//2]
    
    #print centering_x
    Shifting(xcDic, ycDic)
    
    #Scale the secondary structure.
    #scaling now structure
    # scaling center coordinates on page
    ry = 300
    scx = 0 ## scaling center coordinates on page
    scy = 0
    
    minx = min([xcDic[key] for key in xcDic.keys()])
    maxx = max([xcDic[key] for key in xcDic.keys()])
    miny = min([ycDic[key] for key in ycDic.keys()])
    maxy = max([ycDic[key] for key in ycDic.keys()])
    
    
    scfactor = ry/(maxy-miny)
    if scfactor < 1:
        for i in range(len(xcDic)):
            tx = xcDic[i] - scx
            ty = ycDic[i] - scy
            xcDic[i] = tx*scfactor + scx
            ycDic[i] = ty*scfactor + scy
    
    # mirror sequence so that when it rotate, the 5' will be on the top.
    for i in range(len(xcDic)):
        xcDic[i] = -xcDic[i]
    Shifting(xcDic, ycDic)
    
    # Determine the rotated angle phi
    # middle part of the mature RNA is the rotation center
    ax = xcDic[centering_x]
    ay = ycDic[centering_x]
     
    # point relativ to center
    if twisted == 1:
        bx = xcDic[centering_x-1]
        by = ycDic[centering_x-1]
    else:
        bx = xcDic[centering_x+1]
        by = ycDic[centering_x+1]
    gk = by-ay
    ak = bx-ax
    r = sqrt((ak**2)+(gk**2))
    phi = asin(gk/r)
    if (bx < ax and by > ay):
        phi = 3.141593-phi
    if (bx <= ax and by <= ay):
        phi = phi*(-1)
        phi == phi + 3.141593
    #This phi is abandoned, so only set phi to be 3.141593*1.5
    phi = 3.141593*1.5
    # Now rotate every point in a designated angle of phi
    for i in range(len(xcDic)):
        if i == centering_x:
            pass
        else:
            bx = xcDic[i]
            by = ycDic[i]
            gk = by-ay
            ak = bx-ax
            r = sqrt((ak**2)+(gk**2))
            alpha = asin(gk/r)
            if bx < ax and by > ay:
                alpha = 3.141593-alpha
            if bx <= ax and by <= ay:
                alpha = alpha*(-1)
                alpha = alpha + 3.141593
            alpha = alpha - phi
            xcDic[i] = ax + r*cos(alpha)
            ycDic[i] = ay + r*sin(alpha)
    Shifting(xcDic, ycDic)
    
    # Translocate the structure to the proper postion
    minx = min([xcDic[key] for key in xcDic.keys()])
    maxx = max([xcDic[key] for key in xcDic.keys()])
    miny = min([ycDic[key] for key in ycDic.keys()])
    maxy = max([ycDic[key] for key in ycDic.keys()])
    cnetroidX, centroidY = centroidOfRectangle(minx, maxx, miny, maxy)
    #print (cnetroidX, centroidY)
    scx = 95
    scy = 380
    tx = scx - cnetroidX
    ty = scy - centroidY
    for i in range(len(xcDic)):
        xcDic[i] = xcDic[i] + tx
        ycDic[i] = ycDic[i] + ty

    # Draw other parts of the pdf
    y = 300 #300;
    x = 300
    c.setLineWidth(0.5)
    c.setStrokeColor(colors.grey)

    # Draw lines
    for i in range(len(xcDic)):
        if i > 0:
            c.line(xcDic[i-1]+2.5+x,ycDic[i-1]+2+y,xcDic[i]+2.5+x,ycDic[i]+2+y)

    # Draw bases
    c.setFont('Courier', 10*scfactor)
    for i in range(len(sequence)):
        if i in matureMiRNALocationList:
            c.setFillColor(colors.red)
        elif i in starMiRNALocationList:
            c.setFillColor(colors.lightskyblue)
        else:
            c.setFillColor(colors.black)
        c.drawString(xcDic[i]+x,ycDic[i]+y,sequence[i])

    # Draw base pairing
    scfactorl= 0.4
    c.setLineWidth(0.6)
    c.setStrokeColor(colors.black)
    for key in bpDic.keys():
        dx = abs(xcDic[key] - xcDic[bpDic[key]])
        dy = abs(ycDic[key] - ycDic[bpDic[key]])

        dx1 = (dx-scfactorl*dx)/2
        dy1 = (dy-scfactorl*dy)/2

        if xcDic[key] > xcDic[bpDic[key]]:
            fx = xcDic[key] - dx1
            tox = xcDic[bpDic[key]] + dx1
        else:
            fx = xcDic[key] + dx1
            tox = xcDic[bpDic[key]] - dx1

        if ycDic[key] > ycDic[bpDic[key]]:
            fy = ycDic[key] - dy1
            toy = ycDic[bpDic[key]] + dy1
        else:
            fy = ycDic[key] + dy1
            toy = ycDic[bpDic[key]] - dy1
        #c.line(fx+2.5+x,fy+2+y,tox+2.5+x,toy+2+y)
        c.line(fx+2.5*scfactor+x,fy+2*scfactor+y,tox+2.5*scfactor+x,toy+2*scfactor+y)
    # Draw 5' and 3'
    c.setFillColor(colors.black)
    c.setFont('Courier', 8)
    c.drawString(xcDic[0]+2.5+x-12,ycDic[0]+2+y-5, "5'")
    c.drawString(xcDic[len(xcDic)-1]+2.5+x-12,ycDic[len(xcDic)-1]+2+y-5, "3'")
    
    # Draw histogram.
    # draw y axis
    multiplier = 3.6 # minimal distance between two letters
    lstruct_multi = (len(matureMiRNAPrecusorSeq)+2)*multiplier
    c.setLineWidth(2.5)
    c.setStrokeColor(colors.black)
    c.line(xposshift+15, downy+550+1, xposshift+15, downy+450-1)
    # draw x axis
    c.setStrokeColor(colors.grey)
    c.line(xposshift+15, downy+450, xposshift+15+lstruct_multi, downy+450)
    splitXScale = len(matureMiRNAPrecusorSeq)//10
    c.setFont('Courier', 5)
    c.setFillColor(colors.black)
    c.setStrokeColor(colors.grey)
    for i in range(1, splitXScale+1):
        c.line(xposshift+15+i*multiplier*10, downy+450, xposshift+15+i*multiplier*10, downy+453)
        c.drawString(xposshift+12+i*multiplier*10, downy+440, str(i)+'0')
    # draw x axis arrow
    c.setLineWidth(2.5)
    c.setStrokeColor(colors.black)
    c.line(xposshift+12+lstruct_multi, downy+453, xposshift+15+lstruct_multi, downy+450)
    c.line(xposshift+15+lstruct_multi, downy+450, xposshift+12+lstruct_multi, downy+447)
    # draw y axis arrow
    #c.line(xposshift+12, downy+557, xposshift+15, downy+560)
    #c.line(xposshift+15, downy+560, xposshift+18, downy+557)
    # add label of x axis and y axis
    c.setFont('Courier', 8)
    c.drawString(xposshift+15+lstruct_multi*0.5-10, downy+430, "Length(bp)")
    c.setFont('Courier', 6)
    c.drawString(xposshift+0, downy+560, "Percentage")
    # draw scale in y axis
    c.setFont('Courier', 6)
    # 1.0
    c.line(xposshift+15-1.0, downy+550, xposshift+18, downy+550)
    c.drawString(xposshift-5, downy+548, "100%")
    # 0.75
    c.line(xposshift+15, downy+525, xposshift+18, downy+525)
    c.drawString(xposshift-3, downy+523, "75%")
    # 0.5
    c.line(xposshift+15, downy+500, xposshift+18, downy+500)
    c.drawString(xposshift-3, downy+498, "50%")
    # 0.25
    c.line(xposshift+15, downy+475, xposshift+18, downy+475)
    c.drawString(xposshift-3, downy+473, "25%")
    # 0
    c.drawString(xposshift+7, downy+448, "0")
    # Draw the detailed reads information in the cluster.
    # Calculate the frequency of the nucleotide on genome for each postion
    totalReadCount = 0
    for item in matureReadContentlist:
        totalReadCount = totalReadCount + int(item[1])
    if passengerReadContentlist != 'None':
        for item in passengerReadContentlist:
            totalReadCount = totalReadCount + int(item[1])
    frequencyList = []
    # Calculate predefined pdf loci for alignment characters
    position_Dic = {}
    counter = 0
    for i in range(200):
        position_Dic.update({i:xposshift+multiplier+15+i*multiplier})
    
    for index, nucleotide in enumerate(matureMiRNAPrecusorSeq):
        countTmp = 0
        for matureRead in matureReadContentlist:
            if matureRead[0][index] == nucleotide:
                countTmp = countTmp + int( matureRead[1])
        if passengerReadContentlist != 'None':
            for passengerRead in passengerReadContentlist:
                if passengerRead[0][index] == nucleotide:
                    countTmp = countTmp + int( passengerRead[1])
        frequencyList.append(float(countTmp)/totalReadCount)
    # Draw histogram with different clolor
    #print matureMiRNALocationList
    for i in range(len(xcDic)):
        # Assign colors
        if i in matureMiRNALocationList:
            c.setStrokeColor(colors.red)
        elif i in starMiRNALocationList:
            c.setStrokeColor(colors.lightskyblue)
        else:
            c.setStrokeColor(colors.black)
        if i == 0:
            c.line(xposshift+15, downy+450, position_Dic[i], downy+450+frequencyList[i]*100)
        else:
            c.line(position_Dic[i-1], downy+450+frequencyList[i-1]*100, position_Dic[i], downy+450+frequencyList[i]*100)
    # Draw alignment between precursor sequence and read sequences
    for i in range(len(matureMiRNAPrecusorSeq)):
        c.setFont('Courier', 5)
        if i in matureMiRNALocationList:
            c.setFillColor(colors.red)
        elif i in starMiRNALocationList:
            c.setFillColor(colors.lightskyblue)
        else:
            c.setFillColor(colors.black)
        c.drawString(position_Dic[i],downy+400,matureMiRNAPrecusorSeq[i])
    c.setFont('Courier', 5)
    c.setFillColor(colors.black)
    c.drawString(xposshift+20+lstruct_multi, downy+400, "-3'")
    c.drawString(xposshift+5, downy+400, "5'-")
    for i, subitem in enumerate(matureMiRNAPrecusorStr):
        c.drawString(position_Dic[i], downy+390, subitem)
    c.setFont('Courier-Bold', 5)
    c.setFillColor(colors.black)
    c.drawString(xposshift+20+ lstruct_multi, downy+390,'reads')
    #c.drawString(xposshift+60+ lstruct_multi, downy+390,'sample')
    c.setFont('Courier', 5)
    c.setFillColor(colors.black)
    minusValue = 10
    lowLimit = downy+75
    if passengerReadContentlist != 'None':
        totalSeqCount = len(matureReadContentlist) + len(passengerReadContentlist)
        totalReadContentlist = matureReadContentlist + passengerReadContentlist
        chunkLabelList = chunkListNew(totalSeqCount)
    else:
        totalSeqCount = len(matureReadContentlist)
        totalReadContentlist = matureReadContentlist
        chunkLabelList = chunkListNew(totalSeqCount)
    for i, chunk in enumerate(chunkLabelList):
        minusValue = 10
        if i == 0:
            for index in chunk:
                for j, subitem in enumerate(totalReadContentlist[index][0]):
                    c.drawString(position_Dic[j], downy+390-minusValue, subitem)
                c.drawString(xposshift+20+ lstruct_multi, downy+390-minusValue,str(totalReadContentlist[index][1]))
                #c.drawString(xposshift+60+ lstruct_multi, downy+390-minusValue,sampleName)
                minusValue = minusValue + 10
        else:
            c.showPage()
            c.setFont('Courier', 5)
            c.setFillColor(colors.black)
            for index in chunk:
                for j, subitem in enumerate(totalReadContentlist[index][0]):
                    c.drawString(position_Dic[j], downy+710-minusValue, subitem)
                c.drawString(xposshift+20+ lstruct_multi, downy+710-minusValue,str(totalReadContentlist[index][1]))
                #c.drawString(xposshift+60+ lstruct_multi, downy+710-minusValue,sampleName)
                minusValue = minusValue + 10
    c.showPage()
    c.save()

def getMiRNAPosition(miRNASeq, RNASequence):
    startPos = RNASequence.find(miRNASeq)
    endPos = startPos + len(miRNASeq)
    return [startPos, endPos]

def getMiRNAStructure(miRNASeq, RNASequence, RNAStructure):
    [startPos, endPos] = getMiRNAPosition(miRNASeq, RNASequence)
    mirnaStruct = RNAStructure[startPos:endPos]
    return mirnaStruct

def removeDash(seq):
    newSeq = ''
    for unit in seq:
        if unit != '-':
            newSeq = newSeq + unit
    return newSeq


def headDashCount2(clusterSeq):
    # clusterSeq is: CTAATACTGCCTGGTAATGATGACGG-
    headDashCount = 0
    for symbol in clusterSeq:
        if symbol == '-':
            headDashCount = headDashCount + 1
        else:
            break
    return headDashCount

def tailDashCount2(clusterSeq):
    # clusterSeq is: CTAATACTGCCTGGTAATGATGACGG-
    tailDashCount = 0
    for symbol in clusterSeq[::-1]:
        if symbol == '-':
            tailDashCount = tailDashCount + 1
        else:
            break
    return tailDashCount

def chunkInto2(arr):
    # length of list must be even and larger than 2
    return [arr[i:i+2] for i in range(0, len(arr), 2)]

def padClusteredList(matureReadContentlistRaw, matureMiRNASeq, matureMiRNAPrecusorSeq, startPos, endPos, strand):
    startTmp = matureMiRNAPrecusorSeq.index(matureMiRNASeq)
    endTmp = matureMiRNAPrecusorSeq.index(matureMiRNASeq) + len(matureMiRNASeq)
    headLen = startTmp
    tailLen = len(matureMiRNAPrecusorSeq) - endTmp
    if strand == '+':
        precursorHeadLocation = int(startPos) - headLen
        precursorTailLocation = int(endPos) + tailLen
    else:
        precursorHeadLocation = int(startPos) - tailLen
        precursorTailLocation = int(endPos) + headLen
    newReadContentList = []
    for readContent in matureReadContentlistRaw:
        readSeq = readContent[0]
        readSeqStart = int(readContent[1])
        readSeqEnd = int(readContent[2])
        readCount = readContent[3]
        if strand == '+':
            readSeqPadded = '-'*(readSeqStart-precursorHeadLocation)+readSeq+'-'*(precursorTailLocation-readSeqEnd)
        else:
            readSeqPadded = '-'*(precursorTailLocation-readSeqEnd)+readSeq+'-'*(readSeqStart-precursorHeadLocation)
        newReadContentList.append((readSeqPadded, readCount))
    return newReadContentList


def write_novel_report(novelmiRNALListFile, featureFile, clusterFile, rnafoldCmdTmp, outputdir2, files):
    dir_tmp = Path((Path(outputdir2).resolve().parents[0]))/(files+'_novel_miRNAs')
    os.mkdir(dir_tmp)
    #samppleName_tmp = '_'.join(novelmiRNALListFile.split('_')[:-3])
    novelmiRNALOriginalList = []
    clusterNameProbabilityDic = {}
    with open(novelmiRNALListFile, 'r') as inf:
        line = inf.readline()
        probabilityIndex = line.strip().split(',').index('probability')
        clusterNameIndex = line.strip().split(',').index('clusterName')
        line = inf.readline()
        while line != '':
            content = line.strip().split(',')
            probability = content[probabilityIndex]
            clusterName = content[clusterNameIndex]
            if clusterName not in novelmiRNALOriginalList:
                novelmiRNALOriginalList.append(clusterName)
                clusterNameProbabilityDic.update({clusterName:probability})
            line = inf.readline()
    # The majority of the paired miRNAs' pruned precursor RNA sequences should be identical. However, the pruned precursor RNA sequences might be slightly different
    # at tail and head part. This situation happens at a quite low probility due to pruning error.
    # Therefore, the paired miRNA's precursor miRNA of the identified novel miRNA should be corrected in order to plot.
    with open(featureFile, 'r') as inf:
        totalContent = inf.readlines()

    with open(featureFile, 'r') as inf:
        line = inf.readline()
        content = line.strip().split('\t')
        clusterNameIndex = content.index('clusterName')
        pruned_precusor_seqIndex = content.index('pruned_precusor_seq')
        pruned_precusor_strIndex = content.index('pruned_precusor_str')
        upstreamDistanceIndex = content.index('upstreamDistance')
        downstreamDistanceIndex = content.index('downstreamDistance')
        armTypeIndex = content.index('armType')
        pair_stateIndex = content.index('pair_state')
        i = 1
        line = inf.readline()
        while line != '':
            content = line.strip().split('\t')
            if content[clusterNameIndex] in novelmiRNALOriginalList:
                armTypeTmp = content[armTypeIndex]
                pair_state = content[pair_stateIndex]
                pruned_precusor_seq = content[pruned_precusor_seqIndex]
                pruned_precusor_str = content[pruned_precusor_strIndex]
                upstreamDistance = content[upstreamDistanceIndex]
                downstreamDistance = content[downstreamDistanceIndex]
                if pair_state == 'Yes':
                    if upstreamDistance != 'None':
                        if int(upstreamDistance) <= 44:
                            correctLineContent = totalContent[i-1].strip().split('\t')
                            if (correctLineContent[pair_stateIndex]) == 'Yes' and (correctLineContent[armTypeIndex] in ['arm5', 'arm3']) and (correctLineContent[armTypeIndex] != armTypeTmp):
                                correctLineContent[pruned_precusor_seqIndex] = pruned_precusor_seq
                                correctLineContent[pruned_precusor_strIndex] = pruned_precusor_str
                            totalContent[i-1] = '\t'.join(correctLineContent)+'\n'
                        elif int(downstreamDistance) <= 44:
                            correctLineContent = totalContent[i+1].strip().split('\t')
                            if (correctLineContent[pair_stateIndex]) == 'Yes' and (correctLineContent[armTypeIndex] in ['arm5', 'arm3']) and (correctLineContent[armTypeIndex] != armTypeTmp):
                                correctLineContent[pruned_precusor_seqIndex] = pruned_precusor_seq
                                correctLineContent[pruned_precusor_strIndex] = pruned_precusor_str
                            totalContent[i+1] = '\t'.join(correctLineContent)+'\n'
                        else:
                            pass
            line = inf.readline()
            i = i + 1

    corrFile = str(Path(outputdir2)/(files + '_corrected.tsv'))
    with open(corrFile, 'w') as outf:
        for t, totalContentTmp in enumerate(totalContent):
            # Correct the armType if it is wrongly alocated because the armType is based on the stableClusterSeq. This sequnce may be too short.
            # So use the clusterSeq to recorrect the armType.
            if t == 0:
                content = totalContentTmp.strip().split('\t')
                clusterNameIndex = content.index('clusterName')
                clusterSeqIndex = content.index('clusterSeq')
                stableClusterSeqIndex = content.index('stableClusterSeq')
                pruned_precusor_seqIndex = content.index('pruned_precusor_seq')
                pruned_precusor_strIndex = content.index('pruned_precusor_str')
                armTypeIndex = content.index('armType')
                outf.write(totalContentTmp)
            else:
                content = totalContentTmp.strip().split('\t')
                clusterSeq = content[clusterSeqIndex]
                clusterSeqNew = ''
                for nucl in clusterSeq:
                    if nucl == 'T':
                        clusterSeqNew = clusterSeqNew + 'U'
                    else:
                        clusterSeqNew = clusterSeqNew + nucl
                stableClusterSeq = content[stableClusterSeqIndex]
                pruned_precusor_seq = content[pruned_precusor_seqIndex]
                pruned_precusor_str = content[pruned_precusor_strIndex]
                clusterSeqStr = getMiRNAStructure(clusterSeqNew, pruned_precusor_seq, pruned_precusor_str)
                pattern = re.compile('\(.*\)')
                if pattern.search(clusterSeqStr) is not None:
                    if content[armTypeIndex] == 'loop':
                        outf.write(totalContentTmp)
                    else:
                        content[armTypeIndex] = 'loop'
                        outf.write('\t'.join(content)+'\n')
                else:
                    outf.write(totalContentTmp)

    clusterNameFeatureDic = {}
    precursorSeqclusterNameDic = {}
    with open(corrFile, 'r') as inf:
        line = inf.readline()
        content = line.strip().split('\t')
        clusterNameIndex = content.index('clusterName')
        seqCountIndex = content.index('seqCount')
        readCountSumIndex = content.index('readCountSum')
        stableClusterSeqIndex = content.index('stableClusterSeq')
        alignedClusterSeqLabel = content.index('alignedClusterSeq')
        headUnstableLengthLabel = content.index('headUnstableLength')
        tailUnstableLengthLabel = content.index('tailUnstableLength')
        precusorSeqIndex = content.index('pruned_precusor_seq')
        precusorStrIndex = content.index('pruned_precusor_str')
        armTypeIndex = content.index('armType')
        line = inf.readline()
        while line != '':
            content = line.strip().split('\t')
            clusterName = content[clusterNameIndex]
            seqCount = content[seqCountIndex]
            readCountSum = content[readCountSumIndex]
            stableClusterSeq = str(Seq(content[stableClusterSeqIndex]).transcribe())
            alignedClusterSeq = content[alignedClusterSeqLabel]
            headUnstableLength = int(content[headUnstableLengthLabel])
            tailUnstableLength = int(content[tailUnstableLengthLabel])
            precusorSeq = content[precusorSeqIndex]
            precusorStr = content[precusorStrIndex]
            armType = content[armTypeIndex]
            strand = clusterName[-1]
            chr = clusterName.split(':')[2]
            startPos = int(clusterName.split(':')[3][:-1].split('_')[0].strip())
            endPos = int(clusterName.split(':')[3][:-1].split('_')[1].strip())
            headDashCountTmp = headDashCount2(alignedClusterSeq)
            tailDashCountTmp = tailDashCount2(alignedClusterSeq)
            if strand == '+':
                startPos = startPos - headDashCountTmp + headUnstableLength
                endPos = endPos + tailDashCountTmp - tailUnstableLength
            else:
                startPos = startPos - tailDashCountTmp + tailUnstableLength
                endPos = endPos + headDashCountTmp - headUnstableLength
            if precusorSeq != 'None':
                if clusterName not in clusterNameFeatureDic.keys():
                    clusterNameFeatureDic.update({clusterName:[chr, startPos, endPos, strand, stableClusterSeq, seqCount, readCountSum, armType, precusorSeq, precusorStr]})
                if (precusorSeq, chr, strand) not in precursorSeqclusterNameDic.keys():
                    precursorSeqclusterNameDic.update({(precusorSeq, chr, strand):[clusterName]})
                else:
                    precursorSeqclusterNameDic[(precusorSeq, chr, strand)].append(clusterName)
            line = inf.readline()

    clusterNameClusterSeqDic = {}
    # Parse the clusterFile to get the detailed reads information for each cluster.
    with open(clusterFile, 'r') as inf:
        contentTmp = inf.readlines()
    labelList = []
    for index, item in enumerate(contentTmp):
        if 'Cluster Name:' in item:
            labelList.append(index)

    for k in range(len(labelList)):
        if k != len(labelList)- 1:
            subContentTmp = contentTmp[labelList[k]:labelList[k+1]]
        else:
            subContentTmp = contentTmp[labelList[k]:]
        clusterNameTmp = subContentTmp[0].strip().split(' ')[2]
        if clusterNameTmp in clusterNameFeatureDic.keys():
            #print clusterNameTmp
            startTmp = int(clusterNameTmp[:-1].split(':')[-1].split('_')[0])
            endTmp = int(clusterNameTmp[:-1].split(':')[-1].split('_')[1])
            if clusterNameTmp[-1] == '+':
                dashedClusterSeqStart = startTmp - headDashCount2(subContentTmp[4].strip())
                dashedClusterSeqEnd = endTmp + tailDashCount2(subContentTmp[4].strip())
            else:
                dashedClusterSeqStart = startTmp - tailDashCount2(subContentTmp[4].strip())
                dashedClusterSeqEnd = endTmp + headDashCount2(subContentTmp[4].strip())
            readContentlist = []
            for subitem in subContentTmp[5:]:
                if clusterNameTmp[-1] == '+':
                    ReadSeqStart = dashedClusterSeqStart + headDashCount2(subitem.split('\t')[0])
                    ReadSeqEnd = dashedClusterSeqEnd - tailDashCount2(subitem.split('\t')[0])
                else:
                    #ReadSeqStart = dashedClusterSeqStart + headDashCount2(subitem.split('\t')[0])-2
                    #ReadSeqEnd = dashedClusterSeqEnd - tailDashCount2(subitem.split('\t')[0])-2
                    
                    ReadSeqStart = dashedClusterSeqStart + tailDashCount2(subitem.split('\t')[0])
                    ReadSeqEnd = dashedClusterSeqEnd - headDashCount2(subitem.split('\t')[0])
                readCount = int(subitem.strip().split('\t')[1])
                #readSeq = removeDash(subitem.split('\t')[0])
                readSeq = str(Seq(removeDash(subitem.split('\t')[0])).transcribe())
                readContentlist.append((readSeq, ReadSeqStart, ReadSeqEnd, readCount))
            clusterNameClusterSeqDic.update({clusterNameTmp:readContentlist})
    
    #print(clusterNameClusterSeqDic)
    #print()
    
    # Output the novel miRNA report csv file
    #print precursorSeqclusterNameDic[('CUGACUGCCGAGGGGGCCCUGGCCUGGAUCCAUGCUGGGCAGAAGCAGCUGGACACUGACCAGGACCCCCCAGGGCCGGAGGAACC', 'chr9', '+')]
    finalFile = str(Path(dir_tmp)/(files+'_novel_miRNAs_report.csv'))
    #finalFile = str(Path(outputdir2)/(files+'_novel_miRNAs_report.csv'))
    html_data = FormatJS(str(Path(outputdir2).resolve().parents[0]))
    #print(str(Path(outputdir2).resolve().parents[0])) files+'_novel_miRNAs'
    #"14","<a href=\"SRR8557389_novel_miRNAs/SRR8557389_novel_miRNA_14.pdf\" target=\"_blank\">SRR8557389_novel_miRNA_14</a>","0.829119033","chr1","26554575","26554595","CUCCUGCCCUCCUUGCUGUAG","29"
    outf1 = open(finalFile, 'w')
    outf1.write('Novel miRNA name,Probability,Chr,Start Pos,End Pos,Strand,Mature miRNA sequence,Arm type,Passenger miRNA sequence,Mature miRNA read Count,Passenger miRNA read Count,Precursor miRNA sequence,Precursor miRNA structure\n')
    i = 1
    while len(novelmiRNALOriginalList) >= 1:
        js_nmirVar= ""
        novelmiRNA = novelmiRNALOriginalList[0]
        sampleName = str(files)
        #sampleName = '_'.join(novelmiRNA.split(':')[0].split('_')[2:])
        precursorSeq = clusterNameFeatureDic[novelmiRNA][-2]
        clusterNameList = precursorSeqclusterNameDic[(precursorSeq, clusterNameFeatureDic[novelmiRNA][0], clusterNameFeatureDic[novelmiRNA][3])]
        for clusterNameListTmp in chunkInto2(clusterNameList):
            flag_padClust = 0
            if len(clusterNameListTmp) == 1:
                matureMiRNAName = clusterNameListTmp[0]
                passengerMiRNAName = 'None'
            elif len(clusterNameListTmp) == 2:
                if clusterNameListTmp[0] in clusterNameProbabilityDic.keys() and clusterNameListTmp[1] in clusterNameProbabilityDic.keys():
                    if int(clusterNameFeatureDic[clusterNameListTmp[0]][6]) >= int(clusterNameFeatureDic[clusterNameListTmp[1]][6]):
                        matureMiRNAName = clusterNameListTmp[0]
                        passengerMiRNAName = clusterNameListTmp[1]
                    else:
                        matureMiRNAName = clusterNameListTmp[1]
                        passengerMiRNAName = clusterNameListTmp[0]
                elif clusterNameListTmp[0] in clusterNameProbabilityDic.keys() and clusterNameListTmp[1] not in clusterNameProbabilityDic.keys():
                    matureMiRNAName = clusterNameListTmp[0]
                    passengerMiRNAName = clusterNameListTmp[1]
                else:
                    matureMiRNAName = clusterNameListTmp[1]
                    passengerMiRNAName = clusterNameListTmp[0]
                if clusterNameFeatureDic[passengerMiRNAName][7] == 'loop':
                    passengerMiRNAName = 'None'
            chr = clusterNameFeatureDic[matureMiRNAName][0]
            startPos = clusterNameFeatureDic[matureMiRNAName][1]
            endPos = clusterNameFeatureDic[matureMiRNAName][2]
            strand = clusterNameFeatureDic[matureMiRNAName][3]
            matureMiRNASeq = clusterNameFeatureDic[matureMiRNAName][4]
            readCountSumMatureMiRNA = clusterNameFeatureDic[matureMiRNAName][6]
            armType = clusterNameFeatureDic[matureMiRNAName][7]
            matureMiRNAPrecusorSeq = clusterNameFeatureDic[matureMiRNAName][8]
            matureMiRNAPrecusorStr = clusterNameFeatureDic[matureMiRNAName][9]
            matureReadContentlistRaw = clusterNameClusterSeqDic[matureMiRNAName]
            if passengerMiRNAName == 'None':
                passengerMiRNASeq = 'None'
                passengerReadContentlistRaw = 'None'
                passengerReadContentlist = 'None'
                readCountSumPassengerMiRNA = 'None'
                passengerStrand = 'None'
            else:
                passengerMiRNASeq = clusterNameFeatureDic[passengerMiRNAName][4]
                readCountSumPassengerMiRNA = clusterNameFeatureDic[passengerMiRNAName][6]
                passengerReadContentlistRaw = clusterNameClusterSeqDic[passengerMiRNAName]
                passengerStartPos = clusterNameFeatureDic[passengerMiRNAName][1]
                passengerEndPos = clusterNameFeatureDic[passengerMiRNAName][2]
                passengerStrand = clusterNameFeatureDic[passengerMiRNAName][3]
            # Pad the head and tail part with '.' according to the allignment to the precursor miRNA.
            try:
                startTmp = matureMiRNAPrecusorSeq.index(matureMiRNASeq)
                flag_padClust = 0
            except ValueError:
                flag_padClust = 1
            
            if flag_padClust == 1:
                continue
            else:
                matureReadContentlist = padClusteredList(matureReadContentlistRaw, matureMiRNASeq, matureMiRNAPrecusorSeq, startPos, endPos, strand)

            if passengerReadContentlistRaw != 'None':
                try:
                    #startTmp = matureMiRNAPrecusorSeq.index(passengerMiRNASeq)
                    passengerReadContentlist = padClusteredList(passengerReadContentlistRaw, passengerMiRNASeq, matureMiRNAPrecusorSeq, passengerStartPos, passengerEndPos, passengerStrand)
                    totalReadCountSum = int(readCountSumMatureMiRNA) + int(readCountSumPassengerMiRNA)
                except ValueError:
                    passengerReadContentlist = 'None'
                    totalReadCountSum = int(readCountSumMatureMiRNA)
            else:
                passengerReadContentlist = 'None'
                totalReadCountSum = int(readCountSumMatureMiRNA)
            if armType != 'loop':
                novelmiRNANameNew = str(files)+'_novel_miRNA_'+str(i)
                try:
                    outf1.write(','.join([novelmiRNANameNew, clusterNameProbabilityDic[matureMiRNAName], chr, str(startPos), str(endPos), strand, matureMiRNASeq, armType, passengerMiRNASeq, str(readCountSumMatureMiRNA), str(readCountSumPassengerMiRNA), matureMiRNAPrecusorSeq, matureMiRNAPrecusorStr]))
                    js_nmirVar = '"<a href=\\"'+str(files)+'_novel_miRNAs/'+str(files)+'_novel_miRNA_'+str(i)+'.pdf\\" target=\\"_blank\\">'+novelmiRNANameNew+'</a>","'+str(clusterNameProbabilityDic[matureMiRNAName])+'","'+chr+'","'+str(startPos)+'","'+str(endPos)+'","'+matureMiRNASeq+'","'+str(readCountSumMatureMiRNA)+'"'
                    html_data.addSerialNum(js_nmirVar)
                except KeyError:
                    outf1.write(','.join([novelmiRNANameNew, "1", chr, str(startPos), str(endPos), strand, matureMiRNASeq, armType, passengerMiRNASeq, str(readCountSumMatureMiRNA), str(readCountSumPassengerMiRNA), matureMiRNAPrecusorSeq, matureMiRNAPrecusorStr]))
                    js_nmirVar = '"<a href=\\"'+str(files)+'_novel_miRNAs/'+str(files)+'_novel_miRNA_'+str(i)+'.pdf\\" target=\\"_blank\\">'+novelmiRNANameNew+'</a>","1","'+chr+'","'+str(startPos)+'","'+str(endPos)+'","'+matureMiRNASeq+'","'+str(readCountSumMatureMiRNA)+'"'
                    html_data.addSerialNum(js_nmirVar)
                outf1.write('\n')
                # Prepare to plot the precurosr sturcuture, cluster seuqences into a pdf file.
                with open(str(Path(outputdir2)/(files+'_precusorTmp.fa')), 'w') as outf:
                    fa_tmp = '\n'.join(['>'+novelmiRNANameNew, matureMiRNAPrecusorSeq, matureMiRNAPrecusorStr])
                    outf.write(fa_tmp+'\n')
                f1 = str(files+'_precusorTmp.fa')
                f2 = str(files+'_precusorTmp.str')
                os.system('cd %s && %s -d 0 < %s > %s'%(Path(outputdir2), rnafoldCmdTmp, f1, f2))
                f3 = str(Path(outputdir2)/(files+'_novel_miRNA_'+str(i)+'_ss.ps'))
                f4 = str(Path(dir_tmp)/(files+'_novel_miRNA_'+str(i)+'.pdf'))
                try:
                    creatPDF(sampleName, novelmiRNANameNew, clusterNameProbabilityDic[matureMiRNAName], chr, startPos, endPos, strand, armType, readCountSumMatureMiRNA, totalReadCountSum, matureMiRNAPrecusorSeq, matureMiRNAPrecusorStr, matureMiRNASeq, passengerMiRNASeq, f3, f4, matureReadContentlist, passengerReadContentlist)
                except KeyError:
                    creatPDF(sampleName, novelmiRNANameNew, "1", chr, startPos, endPos, strand, armType, readCountSumMatureMiRNA, totalReadCountSum, matureMiRNAPrecusorSeq, matureMiRNAPrecusorStr, matureMiRNASeq, passengerMiRNASeq, f3, f4, matureReadContentlist, passengerReadContentlist)
                
                i = i + 1
            # Delete the clusterNames in clusterNameListTmp
            for clusterName in clusterNameListTmp:
                #print '%s is remvoed'%(clusterName)
                if clusterName in novelmiRNALOriginalList:
                    novelmiRNALOriginalList.remove(clusterName)     
    outf1.close()
